# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
from moleculekit.molecule import Molecule
import numpy as np
import logging

logger = logging.getLogger(__name__)


def _segid_gen(prefix, mol, mode="decimal"):
    import string

    segids = np.unique(mol.segid)
    for prefix in [prefix] + list(string.ascii_uppercase + string.digits):
        i = 0
        while True:
            if mode == "decimal":
                segid = f"{prefix}{i:d}"
            elif mode == "hex":
                segid = f"{prefix}{i:X}"
            elif mode == "alphanum":
                segid = "{0}{1:c}{2:c}{3:c}".format(
                    prefix,
                    int(np.floor(np.floor(i / 26) / 26) + 65),
                    int(np.mod(np.floor(i / 26), 26) + 65),
                    int(np.mod(i, 26) + 65),
                )
            if len(segid) > 4:
                break
            i += 1

            if segid not in segids:
                yield segid


def solvate(
    mol,
    pad=None,
    minmax=None,
    centersel=None,
    boxsize=None,
    negx=0,
    posx=0,
    negy=0,
    posy=0,
    negz=0,
    posz=0,
    buffer=2.4,
    watsize=65.4195,
    prefix="W",
    rotate=False,
    spdb=None,
):
    """Solvates the system in a water box


    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The molecule object we want to solvate
    pad : float
        The padding to add to the minmax in all dimensions. You can specify different padding in each dimension using
        the negx, negy, negz, posx, posy, posz options. This option will override any values in the neg and pos options.
    minmax : list
        Min and max dimensions. Should be a 2D matrix of the form [[minx, miny, minz], [maxx, maxy, maxz]]. If none is
        given, it is calculated from the minimum and maximum coordinates in the mol.
    centersel : str
        An atom selection string defining the center of the solvation box. The geometric center of the selected atoms is
        used. Must be combined with `boxsize`.
    boxsize : float or list of 3 floats
        The dimensions of the solvation box. If a single float, a cubic box of that side length is created. If a
        3-element iterable [sx, sy, sz], an axis-aligned box of those dimensions is created. The box extends from
        center - boxsize/2 to center + boxsize/2 in each dimension. Must be combined with `centersel`.
    negx : float
        The padding in the -x dimension
    posx : float
        The padding in the +x dimension
    negy : float
        The padding in the -y dimension
    posy : float
        The padding in the +y dimension
    negz : float
        The padding in the -z dimension
    posz : float
        The padding in the +z dimension
    buffer : float
        How much buffer space to leave empty between waters and other molecules
    watsize : float
        The size of the water box
    prefix : str
        The prefix used for water segments
    keysel : str
        The key selection for water atoms
    rotate : bool
        Enable automated rotation of molecule to fit best in box
    rotsel : str
        The selection of atoms to rotate
    rotinc : float
        The increment in degrees to rotate
    spdb : str
        The path to the water pdb file

    Returns
    -------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A solvated molecule

    Examples
    --------
    >>> smol = solvate(mol, pad=10)
    >>> smol = solvate(mol, minmax=[[-20, -20, -20],[20, 20, 20]])
    >>> smol = solvate(mol, centersel="protein", boxsize=100)
    >>> smol = solvate(mol, centersel="protein", boxsize=[80, 80, 120])
    """
    from tqdm import tqdm
    from htmd.home import home

    mol = mol.copy()
    if mol.numFrames > 1:
        logger.warning(
            "Multiple frames in Molecule. Solvate keeps only frame 0 and discards the rest."
        )
        mol.coords = np.atleast_3d(mol.coords[:, :, 0])

    if (centersel is None) != (boxsize is None):
        raise ValueError("centersel and boxsize must both be specified together.")

    if centersel is not None:
        if minmax is not None or pad is not None:
            raise ValueError("centersel/boxsize cannot be combined with minmax or pad.")
        selatoms = mol.atomselect(centersel)
        if not np.any(selatoms):
            raise ValueError(f"Atom selection '{centersel}' matched no atoms.")
        center = mol.get("coords", sel=selatoms).mean(axis=0)
        boxsize = np.atleast_1d(np.array(boxsize, dtype=float))
        if boxsize.shape == (1,):
            boxsize = np.repeat(boxsize, 3)
        elif boxsize.shape != (3,):
            raise ValueError("boxsize must be a scalar or a 3-element iterable.")
        half = boxsize / 2.0
        minmax = np.array([center - half, center + half])
        logger.info(
            f"Box center from selection '{centersel}': [{center[0]:.2f}, {center[1]:.2f}, {center[2]:.2f}], "
            f"box size: [{boxsize[0]:.2f}, {boxsize[1]:.2f}, {boxsize[2]:.2f}]"
        )

    if spdb is None:
        spdb = os.path.join(home(shareDir=True), "solvate", "wat.pdb")

    if os.path.isfile(spdb):
        logger.info("Using water pdb file at: " + spdb)
        water = Molecule(spdb)
    else:
        raise NameError("No solvent pdb file found in " + spdb)

    if pad is not None:
        negx = pad
        posx = pad
        negy = pad
        posy = pad
        negz = pad
        posz = pad

    if rotate:
        raise NameError("Rotation not implemented yet")

    # Calculate min max coordinates from molecule
    if mol.numAtoms > 0:
        minmol = np.min(mol.get("coords"), axis=0)
        maxmol = np.max(mol.get("coords"), axis=0)
    else:
        minmol = [np.inf, np.inf, np.inf]
        maxmol = [-np.inf, -np.inf, -np.inf]

    if minmax is None:
        minc = minmol
        maxc = maxmol
    else:
        if isinstance(minmax, list):
            minmax = np.array(minmax)
        minc = minmax[0, :]
        maxc = minmax[1, :]

    xmin = float(minc[0] - negx)
    xmax = float(maxc[0] + posx)
    ymin = float(minc[1] - negy)
    ymax = float(maxc[1] + posy)
    zmin = float(minc[2] - negz)
    zmax = float(maxc[2] + posz)

    dx = xmax - xmin
    dy = ymax - ymin
    dz = zmax - zmin

    nx = int(np.ceil((dx + 2 * buffer) / watsize))
    ny = int(np.ceil((dy + 2 * buffer) / watsize))
    nz = int(np.ceil((dz + 2 * buffer) / watsize))

    # Calculate number of preexisting water segments with given prefix
    if mol.numAtoms > 0:
        preexist = len(np.unique(mol.get("segid", sel=f'segid "{prefix}.*"')))
    else:
        preexist = 0

    numsegs = nx * ny * nz
    logger.info(f"Replicating {numsegs} water segments, {nx} by {ny} by {nz}")

    # Check that we won't run out of segment name characters, and switch to
    # using hexadecimal or alphanumeric naming schemes in cases where decimal
    # numbered segnames won't fit into the field width.
    testsegname = f"{prefix}{numsegs + preexist:d}"
    testsegnamehex = f"{prefix}{numsegs + preexist:X}"
    writemode = "decimal"
    if len(testsegname) > 4 and len(testsegnamehex) <= 4:
        writemode = "hex"
        logger.warning(
            "Warning: decimal naming would overrun segname field. Using hexadecimal segnames instead..."
        )
    elif len(testsegnamehex) > 4:
        writemode = "alphanum"
        logger.warning(
            "Warning: decimal or hex naming would overrun segname field. Using alphanumeric segnames instead..."
        )

    segid_gen = _segid_gen(prefix, mol, writemode)

    minx = minmol[0] - buffer
    miny = minmol[1] - buffer
    minz = minmol[2] - buffer
    maxx = maxmol[0] + buffer
    maxy = maxmol[1] + buffer
    maxz = maxmol[2] + buffer

    bar = tqdm(total=nx * ny * nz, desc="Solvating")
    waterboxes = np.empty(numsegs, dtype=object)
    n = preexist
    w = 0
    for i in range(nx):
        movex = xmin + i * watsize
        movexmax = movex + watsize
        xoverlap = True
        if movex > maxx or movexmax < minx:
            xoverlap = False

        for j in range(ny):
            movey = ymin + j * watsize
            moveymax = movey + watsize
            yoverlap = True
            if movey > maxy or moveymax < miny:
                yoverlap = False

            for k in range(nz):
                movez = zmin + k * watsize
                movezmax = movez + watsize
                zoverlap = True
                if movez > maxz or movezmax < minz:
                    zoverlap = False

                segname = next(segid_gen)

                waterboxes[w] = water.copy()
                waterboxes[w].moveBy([movex, movey, movez])
                waterboxes[w].set("segid", segname)

                mol.append(waterboxes[w])
                watsel = mol.segid == segname

                selover = np.zeros(len(watsel), dtype=bool)
                if (
                    xoverlap and yoverlap and zoverlap
                ):  # Remove water overlapping with other segids
                    selover = _overlapWithOther(mol, segname, buffer)
                # Remove water outside the boundaries
                selout = _outOfBoundaries(
                    mol, segname, xmin, xmax, ymin, ymax, zmin, zmax
                )
                sel = selover | selout

                # mol.write('temp.pdb')
                mol.filter(mol.segid != segname, _logger=False)
                waterboxes[w].filter(np.invert(sel[watsel]), _logger=False)
                # waterboxes[w].write('wat' + str(w) + '.pdb')
                n += 1
                w += 1
                bar.update(1)
    bar.close()

    waters = 0
    for i in range(numsegs):
        waters += waterboxes[i].numAtoms
        if waterboxes[i].numAtoms != 0:
            mol.append(waterboxes[i])

    logger.info(f"{int(waters / 3)} water molecules were added to the system.")
    return mol


def _overlapWithOther(mol, segname, buffer):
    # Optimized version of this atomselection:
    # segid {segname} and same resid as (segid {segname} and within {buffer} of not segid {segname})
    from moleculekit.atomselect_utils import within_distance

    segmask = mol.segid == segname
    segidx = np.where(segmask)[0].astype(np.uint32)
    notsegidx = np.where(~segmask)[0].astype(np.uint32)

    contacts = np.zeros(len(segidx), dtype=bool)
    within_distance(
        mol.coords[:, :, 0],
        buffer,
        sel1=segidx,
        sel2=notsegidx,
        sel2_min_coords=mol.coords[notsegidx, :, 0].min(axis=0),
        sel2_max_coords=mol.coords[notsegidx, :, 0].max(axis=0),
        results=contacts,
    )

    close_resid = np.unique(mol.resid[segidx[contacts]])
    res = segmask & np.isin(mol.resid, close_resid)
    return res


def _outOfBoundaries(mol, segname, xmin, xmax, ymin, ymax, zmin, zmax):
    # Implementing the following atomselection
    # segid {segname} and same resid as (segid {segname} and (x < {xmin} or x > {xmax} or y < {ymin} or y > {ymax} or z < {zmin} or z > {zmax}))

    segnamesel = mol.segid == segname
    oob = (
        (mol.x < xmin)
        | (mol.x > xmax)
        | (mol.y < ymin)
        | (mol.y > ymax)
        | (mol.z < zmin)
        | (mol.z > zmax)
    )
    residsel = np.isin(mol.resid, mol.resid[segnamesel & oob])
    return segnamesel & residsel



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.tools.graphalignment import makeMolGraph, compareGraphs
from moleculekit.molecule import Molecule
from moleculekit.util import ensurelist
from htmd.home import home

import networkx as nx

import shutil
import parmed
import numpy as np
import tempfile
import os
import logging

logger = logging.getLogger(__name__)


triala = Molecule(os.path.join(home(shareDir=True), "builder", "triala_capped.cif"))
triala_g = triala.toGraph()
triala_bb = nx.shortest_path(triala_g, source=1, target=38)
triala_g.remove_edge(14, 16)
triala_g.remove_edge(24, 26)
triala_comp = sorted(
    [list(x) for x in nx.connected_components(triala_g)],
    key=lambda c: min(c),
)
start_map = np.zeros(triala.numAtoms, dtype=bool)
start_map[triala_comp[0]] = True
end_map = np.zeros(triala.numAtoms, dtype=bool)
end_map[triala_comp[2]] = True


def _extend_residue(mol, nterm=True, cterm=True):
    # Adds ALA on either side of the residue for full parameterization
    mol = mol.copy()
    mol.resid[:] = 3

    def _ix(mol, name, resid=3):
        return np.where((mol.name == name) & (mol.resid == resid))[0][0]

    backbone_idx = nx.shortest_path(
        mol.toGraph(), source=_ix(mol, "N"), target=_ix(mol, "C")
    )
    if nterm:
        # Remove backbone formal charge since we'll add ALA
        mol.formalcharge[_ix(mol, "N")] = 0
        try:
            hn2_idx = _ix(mol, "HN2")
        except Exception:
            logger.info(
                "Could not find HN2 atom. Will create peptide bond using H atom of backbone N."
            )
            hn2_idx = _ix(mol, "H")

        startcap = triala.copy()
        startcap.align([14, 16, 18], refmol=mol, refsel=[hn2_idx] + backbone_idx[:2])
        startcap.filter(start_map, _logger=False)

        mol.remove(f"index {hn2_idx} or name HN2", _logger=False)
        mol.insert(startcap, index=0)
        mol.bonds = np.vstack((mol.bonds, [14, _ix(mol, "N")]))
        mol.bondtype = np.hstack((mol.bondtype, ["1"]))

    backbone_idx = nx.shortest_path(
        mol.toGraph(), source=_ix(mol, "N"), target=_ix(mol, "C")
    )
    if cterm:
        # Remove backbone formal charge since we'll add ALA
        mol.formalcharge[_ix(mol, "C")] = 0
        oxt_idx = _ix(mol, "OXT")

        endcap = triala.copy()
        endcap.align([18, 24, 26], refmol=mol, refsel=backbone_idx[-2:] + [oxt_idx])
        endcap.filter(end_map, _logger=False)

        mol.remove("name OXT HXT", _logger=False)
        mol.append(endcap)
        mol.bonds = np.vstack((mol.bonds, [_ix(mol, "C", 3), _ix(mol, "N", 4)]))
        mol.bondtype = np.hstack((mol.bondtype, ["1"]))

    # TODO: Need to add debumping code
    return mol


_CHARGE_METHOD_MAP = {
    "am1-bcc": "bcc",
    "gasteiger": "gas",
    None: None,
    "none": None,
}


def _write_pyodide_output(path, data):
    """Write bytes or str data returned by antechamber_pyodide to a file."""
    with open(path, "wb") as f:
        f.write(data if isinstance(data, bytes) else data.encode())


def _find_antechamber():
    """Return path to native antechamber binary, or None."""
    return shutil.which("antechamber", mode=os.X_OK)


def _run_ambertools(program, cmd, cwd, use_pyodide=False, input_files=None):
    """Run an AmberTools command via native subprocess or antechamber_pyodide.

    Parameters
    ----------
    program : str
        The program name (e.g. "antechamber", "parmchk2", "prepgen").
    cmd : list of str
        Full command-line arguments (program name first).
    cwd : str
        Working directory. Native runs with cwd; pyodide ignores it.
    use_pyodide : bool
        If True, dispatch via antechamber_pyodide.run.
    input_files : dict or None
        For pyodide: mapping of filename -> bytes content.

    Returns
    -------
    dict or None
        Pyodide returns a dict of output filename -> bytes.
        Native returns None (outputs are written to disk).
    """
    if use_pyodide:
        from antechamber_pyodide import run as _ac_run

        return _ac_run(program, cmd, input_files=input_files or {})
    else:
        import subprocess

        result = subprocess.run(cmd, capture_output=True, text=True, cwd=cwd)
        if result.returncode != 0:
            raise RuntimeError(
                f"{program} failed (exit {result.returncode}):\n{result.stderr}"
            )
        return None


def _fftype_antechamber(
    mol,
    tmpdir,
    method="gaff2",
    netcharge=0,
    charge_method=None,
    use_pyodide=False,
    output_mol2="typed.mol2",
    output_frcmod="mol.frcmod",
):
    """Assign GAFF2 atom types and generate frcmod using antechamber/parmchk2.

    Parameters
    ----------
    mol : Molecule
        Input molecule.
    tmpdir : str
        Working directory for intermediate files.
    method : str
        Force-field atom-type method ("gaff" or "gaff2").
    netcharge : int
        Net formal charge of the molecule.
    charge_method : str or None
        Charge model passed to antechamber's ``-c`` flag.
    use_pyodide : bool
        Dispatch via antechamber_pyodide instead of subprocess.
    output_mol2 : str
        Basename for the typed mol2 file written into *tmpdir*.
    output_frcmod : str
        Basename for the frcmod file written into *tmpdir*.

    Returns
    -------
    typed_mol : Molecule
        Molecule with GAFF2 atom types assigned.
    frcmod_path : str
        Absolute path to the generated frcmod file.
    """
    input_mol2 = "input.mol2"
    mol2path = os.path.join(tmpdir, input_mol2)
    mol.write(mol2path)

    ac_charge = _CHARGE_METHOD_MAP.get(charge_method.lower() if charge_method else None)

    cmd = [
        "antechamber",
        "-i",
        input_mol2,
        "-fi",
        "mol2",
        "-o",
        output_mol2,
        "-fo",
        "mol2",
        "-at",
        method.lower(),
        "-nc",
        str(netcharge),
        "-dr",
        "n",
        "-j",
        "1",
    ]
    if ac_charge is not None:
        cmd += ["-c", ac_charge]

    input_files = None
    if use_pyodide:
        with open(mol2path, "rb") as f:
            input_files = {input_mol2: f.read()}

    result = _run_ambertools(
        "antechamber",
        cmd,
        cwd=tmpdir,
        use_pyodide=use_pyodide,
        input_files=input_files,
    )

    typed_path = os.path.join(tmpdir, output_mol2)
    frcmod_path = os.path.join(tmpdir, output_frcmod)

    if use_pyodide:
        typed_mol2_data = result[output_mol2]
        _write_pyodide_output(typed_path, typed_mol2_data)
        input_files = {output_mol2: typed_mol2_data}
    else:
        with open(typed_path, "rb") as f:
            input_files = {output_mol2: f.read()}

    cmd = [
        "parmchk2",
        "-i",
        output_mol2,
        "-f",
        "mol2",
        "-o",
        output_frcmod,
        "-s",
        {"gaff": "1", "gaff2": "2"}[method.lower()],
        "-a",
        "Y",
    ]

    result = _run_ambertools(
        "parmchk2",
        cmd,
        cwd=tmpdir,
        use_pyodide=use_pyodide,
        input_files=input_files,
    )

    if use_pyodide:
        _write_pyodide_output(frcmod_path, result[output_frcmod])

    typed_mol = Molecule(typed_path)
    return typed_mol, frcmod_path


def parameterizeNonCanonicalResidues(
    cifs,
    outdir,
    forcefield="GAFF2",
    charge_model="AM1-BCC",
    is_nterm=False,
    is_cterm=False,
    backend=None,
):
    """Parameterize non-canonical residues.

    Parameters
    ----------
    cifs : str or list of str
        Path to the CIF file(s) containing the non-canonical residues.
    outdir : str
        Path to the output directory.
    forcefield : str, optional
        Force field to use for parameterization. Default is "GAFF2".
    charge_model : str, optional, choices: ["AM1-BCC", "Gasteiger", "Espaloma", "ESP", "RESP-PSI4", "None"]
        Charge model to use for parameterization
    is_nterm : bool, optional
        Whether the residue is a terminal N-term
    is_cterm : bool, optional
        Whether the residue is a terminal C-term
    backend : str or None, optional
        Force a specific backend: "parameterize", "antechamber_pyodide", or
        "antechamber_native".  When None (default) the first available backend
        is selected automatically.

    """
    cifs = ensurelist(cifs)
    if forcefield.lower() not in ("sage", "gaff2"):
        raise AttributeError(
            "Parameterization can only be performed with SAGE or GAFF2 forcefields"
        )

    for cif in cifs:
        mol = Molecule(cif)
        _parameterize_non_canonical_residue(
            mol,
            outdir,
            forcefield,
            is_nterm=is_nterm,
            is_cterm=is_cterm,
            charge_model=charge_model,
            backend=backend,
        )


def _parameterize_non_canonical_residue(
    mol,
    outdir,
    forcefield,
    charge_model,
    is_nterm=False,
    is_cterm=False,
    backend=None,
):
    if backend is not None:
        _backend = backend
    else:
        try:
            from parameterize.main import parameterize_molecule  # noqa: F401

            _backend = "parameterize"
        except ImportError:
            try:
                import antechamber_pyodide  # noqa: F401

                _backend = "antechamber_pyodide"
            except ImportError:
                if _find_antechamber() is not None:
                    _backend = "antechamber_native"
                else:
                    raise ImportError(
                        "No parameterization backend available. Install one of: "
                        "parameterize (private, contact info@acellera.com), "
                        "antechamber_pyodide (for Pyodide environments), "
                        "or AmberTools (conda install ambertools -c conda-forge) "
                        "which provides the antechamber command."
                    )

    os.makedirs(outdir, exist_ok=True)

    if len(np.intersect1d(mol.resname, triala.resname)):
        raise RuntimeError(
            f"Your residue cannot be called any of the following: {' '.join(np.unique(triala.resname))}"
        )

    mol = mol.copy()
    resn = mol.resname[0]

    with tempfile.TemporaryDirectory() as tmpdir:
        xmol = _extend_residue(mol, nterm=not is_nterm, cterm=not is_cterm)

        cmol = xmol.copy()
        cmol.resname[:] = resn
        cmol.resid[:] = 1

        if _backend == "parameterize":
            from parameterize.main import parameterize_molecule

            exclude_atoms = list(np.where(xmol.resname != resn)[0])
            cmol.write(os.path.join(tmpdir, f"{resn}.cif"))

            parameterize_molecule(
                outdir=tmpdir,
                molfile=os.path.join(tmpdir, f"{resn}.cif"),
                forcefield=forcefield,
                minimizer="ffeval",
                charge_type=charge_model,
                keep_atomnames=True,
                exclude_atoms=exclude_atoms,
            )

            typed_mol = Molecule(os.path.join(tmpdir, "parameters", f"{resn}-orig.cif"))
            frcmod = os.path.join(tmpdir, "parameters", f"{resn}.frcmod")
        else:
            netcharge = int(np.sum(cmol.formalcharge))
            typed_mol, frcmod = _fftype_antechamber(
                cmol,
                tmpdir,
                method=forcefield,
                netcharge=netcharge,
                charge_method=charge_model,
                use_pyodide=_backend == "antechamber_pyodide",
            )

        shutil.copy(frcmod, outdir)
        use_pyodide = _backend == "antechamber_pyodide"
        _post_process_parameterize(
            outdir,
            typed_mol,
            frcmod,
            resn,
            refmol=xmol,
            use_pyodide=use_pyodide,
        )


def _post_process_parameterize(
    outdir, mol, frcmod, resn, refmol=None, use_pyodide=False
):
    from collections import defaultdict

    padding_atoms = np.zeros(mol.numAtoms, dtype=bool)
    if refmol is not None:
        # TODO: Move this to parameterize (?)
        fields = ("element",)
        g1 = makeMolGraph(refmol, "all", fields)
        g2 = makeMolGraph(mol, "all", fields)
        _, _, matching = compareGraphs(
            g1, g2, fields=fields, tolerance=0.5, returnmatching=True
        )
        for pp in matching:  # Rename atoms in reference molecule
            mol.name[pp[0]] = refmol.name[pp[1]]
            mol.formalcharge[pp[0]] = refmol.formalcharge[pp[1]]
            mol.resname[pp[0]] = refmol.resname[pp[1]]

        # Antechamber can change the case of atom names (e.g. "CL" -> "Cl").
        # The graph matching above may not cover every atom, so fix any
        # remaining case mismatches using the reference molecule's names.
        ref_name_case = {n.upper(): n for n in np.unique(refmol.name)}
        for i in range(mol.numAtoms):
            upper = mol.name[i].upper()
            if upper in ref_name_case and mol.name[i] != ref_name_case[upper]:
                mol.name[i] = ref_name_case[upper]

        padding_atoms[mol.resname != resn] = True

    # Rename backbone atom types
    original = defaultdict(list)
    backbone_at = {"N": "N", "H": "H", "CA": "CT", "HA": "H1", "C": "C", "O": "O"}
    for key, val in backbone_at.items():
        sel = mol.name == key
        for at in mol.atomtype[sel]:
            original[at].append(val)
        mol.atomtype[sel] = val

    prm = parmed.amber.AmberParameterSet(frcmod)

    # Duplicate parameters for renamed atoms which don't exist in the backbone
    _duplicate_parameters(prm, original)

    # Remove unused parameters
    _clean_prm(prm, mol, list(backbone_at.values()), padding_atoms)
    prm.write(os.path.join(outdir, f"{resn}.frcmod"))

    # Remove the caps
    mol.remove(padding_atoms, _logger=False)

    with tempfile.TemporaryDirectory() as tmpdir:
        mol2file = os.path.join(tmpdir, f"{resn}.mol2")
        mol.write(mol2file)

        acfile = os.path.join(tmpdir, f"{resn}_mod.ac")
        ac_cmd = [
            "antechamber",
            "-i",
            f"{resn}.mol2",
            "-fi",
            "mol2",
            "-o",
            f"{resn}_mod.ac",
            "-fo",
            "ac",
            "-nc",
            str(mol.formalcharge.sum()),
            "-pf",
            "y",
            "-dr",
            "n",
            "-j",
            "0",
            "-an",
            "n",
        ]

        input_files = None
        if use_pyodide:
            with open(mol2file, "rb") as f:
                input_files = {f"{resn}.mol2": f.read()}

        result = _run_ambertools(
            "antechamber",
            ac_cmd,
            cwd=tmpdir,
            use_pyodide=use_pyodide,
            input_files=input_files,
        )
        if use_pyodide:
            _write_pyodide_output(acfile, result[f"{resn}_mod.ac"])

        mol_g = mol.toGraph()
        n_idx = np.where(mol.name == "N")[0][0]
        c_idx = np.where(mol.name == "C")[0][0]
        backbone = nx.shortest_path(mol_g, source=n_idx, target=c_idx)

        mainchain = os.path.join(tmpdir, f"mainchain.{resn.lower()}")
        with open(mainchain, "w") as f:
            f.write("HEAD_NAME N\n")
            f.write("TAIL_NAME C\n")
            for bb_idx in backbone[1:-1]:
                f.write(f"MAIN_CHAIN {mol.name[bb_idx]}\n")
            f.write("PRE_HEAD_TYPE C\n")
            f.write("POST_TAIL_TYPE N\n")
            f.write(f"CHARGE {mol.formalcharge.sum():.1f}\n")

        prepi = os.path.join(outdir, f"{resn}.prepi")
        pg_cmd = [
            "prepgen",
            "-i",
            f"{resn}_mod.ac",
            "-o",
            f"{resn}.prepi",
            "-f",
            "prepi",
            "-m",
            f"mainchain.{resn.lower()}",
            "-rn",
            resn,
        ]

        input_files = None
        if use_pyodide:
            with open(acfile, "rb") as f:
                ac_bytes = f.read()
            with open(mainchain, "rb") as f:
                mc_bytes = f.read()
            input_files = {
                f"{resn}_mod.ac": ac_bytes,
                f"mainchain.{resn.lower()}": mc_bytes,
            }

        result = _run_ambertools(
            "prepgen",
            pg_cmd,
            cwd=tmpdir,
            use_pyodide=use_pyodide,
            input_files=input_files,
        )
        if use_pyodide:
            _write_pyodide_output(prepi, result[f"{resn}.prepi"])
        else:
            shutil.copy(os.path.join(tmpdir, f"{resn}.prepi"), prepi)

        _fix_prepi_atomname_capitalization(mol, prepi)


def _fix_prepi_atomname_capitalization(mol, prepi):
    # Fix wrong prepi atom name capitalization
    uqnames = {x.upper(): x for x in np.unique(mol.name)}

    with open(prepi, "r") as f:
        lines = f.readlines()

    section = None
    for i in range(len(lines)):
        if lines[i].strip() == "":
            section = None

        if section == "atoms":
            if len(lines[i].strip().split()) < 4:
                continue
            old_name = lines[i][6:9].strip()
            # Fix wrong prepi atom name capitalization
            if old_name.upper() in uqnames and uqnames[old_name.upper()] != old_name:
                logger.info(
                    f"Fixed residue {mol.resname[0]} atom name {old_name} -> {uqnames[old_name.upper()]}"
                    " to match the input structure."
                )
                lines[i] = (
                    f"{lines[i][:6]}{uqnames[old_name.upper()]:4s}{lines[i][10:]}"
                )
            continue
        if section in ("loop", "improper"):
            n_pieces = len(lines[i].strip().split())
            for j in range(n_pieces):
                piece = lines[i][j * 5 : (j + 1) * 5].strip()
                if piece.upper() in uqnames and uqnames[piece.upper()] != piece:
                    logger.info(
                        f"Fixed residue {mol.resname[0]} atom name {piece} -> {uqnames[piece.upper()]}"
                        " to match the input structure."
                    )
                    lines[i] = (
                        f"{lines[i][:j*5]}{uqnames[piece.upper()] : >5}{lines[i][(j+1)*5:]}"
                    )
            continue

        if lines[i].strip().startswith("CORR"):
            section = "atoms"
        if lines[i].strip().startswith("LOOP"):
            section = "loop"
        if lines[i].strip().startswith("IMPROPER"):
            section = "improper"

    with open(prepi, "w") as f:
        for line in lines:
            f.write(line)


def _clean_prm(prm, mol, backbone_at, padding_atoms):
    from moleculekit.util import calculateAnglesAndDihedrals

    all_at = np.unique(mol.atomtype[~padding_atoms]).tolist() + backbone_at
    angles, dihedrals = calculateAnglesAndDihedrals(mol.bonds)

    at_dict = {
        "bond_types": mol.atomtype[mol.bonds].tolist(),
        "angle_types": mol.atomtype[angles].tolist(),
        "dihedral_types": mol.atomtype[dihedrals].tolist(),
    }

    to_delete = []
    for at in prm.atom_types:
        if at not in all_at:
            to_delete.append(at)
    for at in to_delete:
        del prm.atom_types[at]

    for param_t in ("bond_types", "angle_types", "dihedral_types"):
        to_delete = []
        seen = []
        for bt in getattr(prm, param_t):
            if (
                not all(np.isin(bt, all_at))
                or all(np.isin(bt, backbone_at))
                or (
                    list(bt) not in at_dict[param_t]
                    and list(bt)[::-1] not in at_dict[param_t]
                )
                or list(bt)[::-1] in seen
            ):
                to_delete.append(bt)
            seen.append(list(bt))

        for bt in to_delete:
            del prm.__dict__[param_t][bt]

    for param_t in ("improper_types", "improper_periodic_types"):
        to_delete = []
        for bt in getattr(prm, param_t):
            if not all(np.isin(bt, all_at)) or all(np.isin(bt, backbone_at)):
                to_delete.append(bt)
        for bt in to_delete:
            del prm.__dict__[param_t][bt]


def _duplicate_parameters(prm, original):
    from copy import deepcopy

    def _gen_permutations(typ, original):
        import itertools

        possibles = []
        for tt in typ:
            if tt in original:
                possibles.append(original[tt] + [tt])
            else:
                possibles.append([tt])
        return list(itertools.product(*possibles))

    # Duplicate parameters for renamed atoms which don't exist in the backbone
    for prm_typ in ("bond_types", "angle_types", "dihedral_types", "improper_types"):
        new_bt = {}
        for typ in getattr(prm, prm_typ):
            perms = _gen_permutations(typ, original)
            if len(perms) == 1:
                continue  # No replacements
            for perm in perms:
                new_bt[perm] = getattr(prm, prm_typ)[typ]

        for bt in new_bt:
            prm.__dict__[prm_typ][bt] = deepcopy(new_bt[bt])



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

import tempfile
import logging
import os
import sys
import numpy as np


logger = logging.getLogger(__name__)


def _getNjobs():
    from htmd.config import _config

    njobs = _config["njobs"]
    if njobs < 0:
        import multiprocessing

        njobs = multiprocessing.cpu_count() + njobs + 1
    return njobs


def tempname(suffix="", create=False):
    if create:
        file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)
    else:
        file = tempfile.NamedTemporaryFile(delete=True, suffix=suffix)
    file.close()
    return file.name


def ensurelist(tocheck, tomod=None):
    """Convert np.ndarray and scalars to lists.

    Lists and tuples are left as is. If a second argument is given,
    the type check is performed on the first argument, and the second argument is converted.
    """
    if tomod is None:
        tomod = tocheck
    if isinstance(tocheck, np.ndarray):
        return list(tomod)
    if isinstance(tocheck, range):
        return list(tocheck)
    if not isinstance(tocheck, list) and not isinstance(tocheck, tuple):
        return [
            tomod,
        ]
    return tomod


def diffMolecules(mol1, mol2, sel=None):
    """Check that name, resname, resid, insertion codes match between two molecules.

    Coordinates are not checked.

    Parameters
    ----------
    mol1 : Molecule
        first structure to compare
    mol2 : Molecule
        second structure to compare
    sel: str
        compare only after filtering to this Atom selection string.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__

    Returns
    -------
    diff: list
        a list of differences, as human-readable strings (empty if structures are equal).

    Examples
    --------
    >>> m=Molecule("3PTB")
    >>> m2=m.copy()
    >>> m2.set("resname","HIE","resid 91")
    >>> diffMolecules(m,m2,sel="name CA")
    ['CA   HIS    91     vs   CA   HIE    91  ']
    """

    from numpy import logical_and as l_and

    diff = []

    m1 = mol1.copy()
    m1.filter(sel)
    m2 = mol2.copy()
    m2.filter(sel)

    eq_name = np.equal(m1.name, m2.name)
    eq_resid = np.equal(m1.resid, m2.resid)
    eq_resname = np.equal(m1.resname, m2.resname)
    eq_insertion = np.equal(m1.insertion, m2.insertion)

    eq_all = l_and(eq_name, l_and(eq_resid, l_and(eq_resname, eq_insertion)))

    neq = np.logical_not(eq_all).nonzero()
    for i in neq[0]:
        diff.append(
            "{:4s} {:4s} {:4d} {:1s}   vs   {:4s} {:4s} {:4d} {:1s}".format(
                m1.name[i],
                m1.resname[i],
                m1.resid[i],
                m1.insertion[i],
                m2.name[i],
                m2.resname[i],
                m2.resid[i],
                m2.insertion[i],
            )
        )
    return diff


def getPdbStrings(mol, sel=None, onlyAtom=True):
    """Return the PDB corresponding to molecule and selection, as a list of strings.

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object
    sel : str
        Atom selection string for what to be outputted.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    onlyAtom : bool
        Only return ATOM/HETATM records (default True)

    Examples
    --------
    >>> m = Molecule("3PTB")
    >>> getPdbStrings(m, "resname BEN")         # doctest: +NORMALIZE_WHITESPACE
    ['HETATM    1  C1  BEN A   1      -1.853  14.311  16.658  1.00 19.86      1    C  ',
     'HETATM    2  C2  BEN A   1      -2.107  15.653  16.758  1.00 19.86      1    C  ',
     'HETATM    3  C3  BEN A   1      -1.774  16.341  17.932  1.00 19.86      1    C  ',
     'HETATM    4  C4  BEN A   1      -1.175  15.662  19.005  1.00 19.86      1    C  ',
     'HETATM    5  C5  BEN A   1      -0.914  14.295  18.885  1.00 19.86      1    C  ',
     'HETATM    6  C6  BEN A   1      -1.257  13.634  17.708  1.00 19.86      1    C  ',
     'HETATM    7  C   BEN A   1      -2.193  13.627  15.496  1.00 19.86      1    C  ',
     'HETATM    8  N1  BEN A   1      -2.797  14.235  14.491  1.00 19.86      1    N  ',
     'HETATM    9  N2  BEN A   1      -1.762  12.391  15.309  1.00 19.86      1    N  ']
    """

    f = tempfile.NamedTemporaryFile(suffix=".pdb")
    mol.write(f.name, sel)
    f.seek(0, 0)
    r = f.read()
    f.close()

    rl = r.decode("ascii").split("\n")

    if onlyAtom:
        rl = [x for x in rl if (x.startswith("ATOM") or x.startswith("HETATM"))]

    return rl


def assertSameAsReferenceDir(compareDir, outdir="."):
    """Check if files in refdir are present in the directory given as second argument AND their content matches.

    Raise an exception if not."""

    import filecmp
    import os

    toCompare = os.listdir(compareDir)
    match, mismatch, error = filecmp.cmpfiles(
        outdir, compareDir, toCompare, shallow=False
    )
    if len(mismatch) != 0 or len(error) != 0 or len(match) != len(toCompare):
        raise Exception(
            f"Different results in files {mismatch}: Compare {outdir} and {compareDir}"
        )


def testDHFR():
    import conda
    import shutil
    from jobqueues.localqueue import LocalGPUQueue

    dhfrdir = os.path.abspath(
        os.path.join(conda.__file__, "../../../../../acemd-examples/dhfr")
    )

    if not os.path.isdir(dhfrdir):
        raise logger.error(
            "Could not find acemd-examples directory. Do `conda install acemd-examples -c acellera`"
        )

    tmpdir = tempname()
    print(tmpdir)
    shutil.copytree(dhfrdir, tmpdir)

    runsh = os.path.join(tmpdir, "run.sh")
    with open(runsh, "w") as f:
        f.write("#!/bin/bash\nacemd >log.txt 2>&1")
    os.chmod(runsh, 0o700)

    logger.info("Starting to run the DHFR test")
    md = LocalGPUQueue()
    try:
        md.submit(tmpdir)
        md.wait()
    except Exception:
        logger.error("Some error occurred. Check {}".format(tmpdir))
    else:
        shutil.rmtree(tmpdir)
        logger.info("Successfully ran the DHRF test. Temporary dir cleaned.")
    return


if __name__ == "__main__":
    from moleculekit.molecule import Molecule
    import doctest

    sys.exit(doctest.testmod().failed)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
import logging

logger = logging.getLogger(__name__)


class KMeansTri(object):
    """Class for calculating the Kmeans-triangle projections of a MetricData  object

    Parameters
    ----------
    data : :class:`MetricData <htmd.metricdata.MetricData>` object
        The object whose data we wish to project.

    Example
    -------
    >>> tri = KMeansTri(data)
    >>> dataproj = tri.project(50)
    """

    def __init__(self, data):
        self.data = data

    def project(self, ndim=None):
        """Projects the data object given to the constructor onto `ndim` dimensions

        Parameters
        ----------
        ndim : int
            The number of dimensions we want to project the data on.

        Returns
        -------
        dataTica : :class:`MetricData <htmd.metricdata.MetricData>` object
            A new :class:`MetricData <htmd.metricdata.MetricData>` object containing the projected data

        Example
        -------
        >>> tri = KMeansTri(data)
        >>> datatri = tri.project(5)
        """
        import scipy.spatial.distance as scidist
        from sklearn.cluster import MiniBatchKMeans
        from htmd.metricdata import MetricData

        datconcat = np.concatenate(self.data.dat)
        mb = MiniBatchKMeans(n_clusters=ndim)
        mb.fit(datconcat)

        # TODO: Could make it into a loop to waste less memory
        dist = scidist.cdist(datconcat, mb.cluster_centers_)
        dist = np.mean(dist, axis=1)[:, np.newaxis] - dist
        dist[dist < 0] = 0

        return MetricData(
            dat=self.data.deconcatenate(dist),
            ref=self.data.ref,
            simlist=self.data.simlist,
            fstep=self.data.fstep,
            parent=self.data,
        )



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from htmd.projections.metric import Metric, _projectionGenerator
from htmd.units import convert as unitconvert
import logging

logger = logging.getLogger(__name__)


class TICA(object):
    """Class for calculating the TICA projections of a MetricData  object

    Time-based Independent Component Analysis
    Projects your data on the slowest coordinates identified for a
    given lagtime.

    Parameters
    ----------
    data : :class:`MetricData <htmd.metricdata.MetricData>` object
        The object whose data we wish to project onto the top TICA dimensions
    lag : int
        The correlation lagtime to use for TICA
    units : str
        The units of lag. Can be 'frames' or any time unit given as a string.
    dimensions : list
        A list of dimensions of the original data on which to apply TICA. All other dimensions will stay unaltered.
        If None is given, it will apply on all dimensions.
    njobs : int
        Number of jobs to spawn for parallel computation of TICA components. If None it will use the default from htmd.config.

    Example
    -------
    >>> from htmd.projections.tica import TICA
    >>> metr = Metric(sims)
    >>> metr.set(MetricSelfDistance('protein and name CA'))
    >>> data = metr.project()
    >>> tica = TICA(data, 20)
    >>> datatica = tica.project(3)
    Alternatively you can pass a Metric object to TICA. Uses less memory but is slower.
    >>> metr = Metric(sims)
    >>> metr.set(MetricSelfDistance('protein and name CA'))
    >>> slowtica = TICA(metr, 20)
    >>> datatica = slowtica.project(3)


    References
    ----------
    Perez-Hernandez, G. and Paul, F. and Giorgino, T. and de Fabritiis, G.
    and Noe, F. (2013) Identification of slow molecular order parameters
    for Markov model construction. J. Chem. Phys., 139 . 015102.
    """

    def __init__(self, data, lag, units="frames", dimensions=None, njobs=None):
        from deeptime.decomposition import TICA as TICAdt
        from tqdm import tqdm
        from htmd.util import _getNjobs

        self.data = data
        self.dimensions = dimensions
        self.njobs = njobs if njobs is not None else _getNjobs()

        if isinstance(
            data, Metric
        ):  # Memory efficient TICA projecting trajectories on the fly
            if units != "frames":
                raise RuntimeError(
                    "Cannot use delayed projection TICA with units other than frames for now. Report this to HTMD issues."
                )
            tic = TICAdt(lagtime=lag)
            metr = data

            pbar = tqdm(total=len(metr.simulations))
            for proj in _projectionGenerator(metr, self.njobs):
                for pro in proj:
                    if pro is None:
                        continue
                    if self.dimensions is None:
                        tic.partial_fit(pro[0])
                    else:  # Sub-select dimensions for fitting
                        tic.partial_fit(pro[0][:, self.dimensions])
                pbar.update(len(proj))
            pbar.close()
        else:  # In-memory TICA
            lag = unitconvert(units, "frames", lag, data.fstep)
            if lag == 0:
                raise RuntimeError(
                    "Lag time conversion resulted in 0 frames. Please use a larger lag-time for TICA."
                )

            tic = TICAdt(lagtime=lag)
            if self.dimensions is None:
                datalist = data.dat
            else:  # Sub-select dimensions for fitting
                datalist = [x[:, self.dimensions].copy() for x in data.dat]
            tic.fit(datalist)
        self.model = tic.fetch_model()

    def project(self, ndim=None, var_cutoff=0.95):
        """Projects the data object given to the constructor onto the top `ndim` TICA dimensions

        Parameters
        ----------
        ndim : int
            The number of TICA dimensions we want to project the data on. If None is given it will use `var_cutoff`
        var_cutoff : float
            Variance cutoff used for automatically determining the number of dimensions

        Returns
        -------
        dataTica : :class:`MetricData <htmd.metricdata.MetricData>` object
            A new :class:`MetricData <htmd.metricdata.MetricData>` object containing the TICA projected data

        Example
        -------
        >>> from htmd.projections.tica import TICA
        >>> tica = TICA(data,20)
        >>> dataTica = tica.project(5)
        """
        from tqdm import tqdm

        if ndim is not None:
            self.model.dim = ndim
            # Replace the following lines in future deeptime versions with self.model.var_cutoff = None
            self.model._var_cutoff = None
            self.model._update_output_dimension()
            # End of hack
        elif var_cutoff is not None:
            self.model.var_cutoff = var_cutoff

        keepdata = []
        keepdim = None
        keepdimdesc = None
        if isinstance(
            self.data, Metric
        ):  # Memory efficient TICA projecting trajectories on the fly
            proj = []
            ref = []
            fstep = None

            metr = self.data
            k = -1
            droppedsims = []
            pbar = tqdm(total=len(metr.simulations))
            for projecteddata in _projectionGenerator(metr, self.njobs):
                for pro in projecteddata:
                    k += 1
                    if pro is None:
                        droppedsims.append(k)
                        continue
                    if self.dimensions is not None:
                        numDimensions = pro[0].shape[1]
                        keepdim = np.setdiff1d(range(numDimensions), self.dimensions)
                        keepdata.append(pro[0][:, keepdim])
                        proj.append(
                            self.model.transform(pro[0][:, self.dimensions]).astype(
                                np.float32
                            )
                        )  # Sub-select dimensions for projecting
                    else:
                        proj.append(self.model.transform(pro[0]).astype(np.float32))
                    ref.append(pro[1])
                    if fstep is None:
                        fstep = pro[2]
                pbar.update(len(projecteddata))
            pbar.close()

            simlist = self.data.simulations
            simlist = np.delete(simlist, droppedsims)
            parent = None
            if self.dimensions is not None:
                from htmd.projections.metric import _singleMolfile
                from moleculekit.molecule import Molecule

                (single, molfile) = _singleMolfile(metr.simulations)
                if single:
                    keepdimdesc = metr.getMapping(Molecule(molfile))
                    keepdimdesc = keepdimdesc.iloc[keepdim]
        else:
            if ndim is not None and self.data.numDimensions < ndim:
                raise RuntimeError(
                    f"TICA cannot increase the dimensionality of your data. Your data has {self.data.numDimensions} dimensions and you requested {ndim} TICA dimensions"
                )

            if self.dimensions is not None:
                keepdim = np.setdiff1d(range(self.data.numDimensions), self.dimensions)
                keepdata = [x[:, keepdim] for x in self.data.dat]
                if self.data.description is not None:
                    keepdimdesc = self.data.description.iloc[keepdim]
            proj = [
                self.model.transform(tr.projection) for tr in self.data.trajectories
            ]
            simlist = self.data.simlist
            ref = self.data.ref
            fstep = self.data.fstep
            parent = self.data

        # If TICA is done on a subset of dimensions, combine non-projected data with projected data
        if self.dimensions is not None:
            newproj = []
            for k, t in zip(keepdata, proj):
                newproj.append(np.hstack((k, t)))
            proj = newproj

        if ndim is None:
            ndim = proj[0].shape[1]
            logger.info(
                f"Kept {ndim} dimension(s) to cover {var_cutoff * 100:.1f}% of kinetic variance."
            )

        from htmd.metricdata import MetricData

        datatica = MetricData(
            dat=proj, simlist=simlist, ref=ref, fstep=fstep, parent=parent
        )
        from pandas import DataFrame

        # TODO: Make this messy pandas creation cleaner. I'm sure I can append rows to DataFrame
        types = []
        indexes = []
        description = []
        for i in range(ndim):
            types += ["tica"]
            indexes += [-1]
            description += [f"TICA dimension {i + 1}"]
        datatica.description = DataFrame(
            {"type": types, "atomIndexes": indexes, "description": description}
        )

        if (
            self.dimensions is not None and keepdimdesc is not None
        ):  # If TICA is done on a subset of dims
            datatica.description = keepdimdesc.append(
                datatica.description, ignore_index=True
            )

        return datatica


if __name__ == "__main__":
    from htmd.simlist import simlist
    from glob import glob
    from moleculekit.projections.metricdistance import MetricSelfDistance
    from htmd.home import home
    from os.path import join

    testfolder = home(dataDir="villin")

    sims = simlist(glob(join(testfolder, "*", "")), join(testfolder, "filtered.pdb"))
    met = Metric(sims[0:2])
    met.set(MetricSelfDistance("protein and name CA"))
    data = met.project()
    data.fstep = 0.1

    tica = TICA(data, 2, dimensions=range(2, 10))
    datatica = tica.project(2)
    tica5 = TICA(data, 0.2, units="ns", dimensions=range(2, 10))
    datatica5 = tica5.project(2)
    expected = [
        [3.69098878, -0.33862674, 0.85779184],
        [3.77816105, -0.31887317, 0.87724227],
        [3.83537507, -0.11878026, 0.65236956],
    ]
    assert np.allclose(
        np.abs(datatica.trajectories[0].projection[-3:, -3:]),
        np.abs(np.array(expected, dtype=np.float32)),
        rtol=0,
        atol=0.01,
    )
    assert np.allclose(
        np.abs(datatica5.trajectories[0].projection[-3:, -3:]),
        np.abs(np.array(expected, dtype=np.float32)),
        rtol=0,
        atol=0.01,
    )
    assert np.all(datatica.description.iloc[[587, 588]].type == "tica")
    assert np.all(datatica.description.iloc[range(587)].type == "distance")
    print("In-memory TICA with subset of dimensions passed test.")

    tica2 = TICA(met, 2, dimensions=range(2, 10))
    datatica2 = tica2.project(2)
    assert np.allclose(
        np.abs(datatica2.trajectories[0].projection[-3:, -3:]),
        np.abs(np.array(expected, dtype=np.float32)),
        rtol=0,
        atol=0.01,
    )
    assert np.all(datatica2.description.iloc[[587, 588]].type == "tica")
    assert np.all(datatica2.description.iloc[range(587)].type == "distance")
    print("Streaming TICA with subset of dimensions passed test.")

    # assert np.max(np.abs(datatica.dat[0][:, -2:]) - np.abs(datatica2.dat[0][:, -2:])) < 0.01, 'Streaming and memory subdim TICA inconsistent.'

    tica3 = TICA(data, 2)
    datatica3 = tica3.project(2)
    expected = [
        [-1.36328638, -0.35354128],
        [-1.35348749, -0.13028328],
        [-1.43249917, -0.31004715],
    ]
    assert np.allclose(
        np.abs(datatica3.trajectories[0].projection[-3:, :]),
        np.abs(np.array(expected, dtype=np.float32)),
        rtol=0,
        atol=0.01,
    )
    assert np.all(datatica3.description.iloc[[0, 1]].type == "tica")
    print("In-memory TICA passed test.")

    tica4 = TICA(met, 2)
    datatica4 = tica4.project(2)
    assert np.allclose(
        np.abs(datatica4.trajectories[0].projection[-3:, :]),
        np.abs(np.array(expected, dtype=np.float32)),
        rtol=0,
        atol=0.01,
    )
    assert np.all(datatica4.description.iloc[[0, 1]].type == "tica")
    print("Streaming TICA passed test.")

    assert (
        np.max(
            np.abs(datatica4.trajectories[0].projection)
            - np.abs(datatica3.trajectories[0].projection)
        )
        < 0.01
    ), "Streaming and memory TICA inconsistent."



"""
HTMD can handle a large amount of simulations.
Simulation lists allow to create a simple list containing all relevant information about the simulations to later
perform any type of analysis.
"""

# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.readers import _TOPOLOGY_READERS
import os
import h5py
import os.path as path
from os import makedirs
from glob import glob as glob
import numpy as np
import unittest
import logging

logger = logging.getLogger(__name__)


class Frame(object):
    """Class used for referencing a specific frame of a :class:`Sim <htmd.simlist.Sim>` object object.

    Parameters
    ----------
    sim : :class:`Sim <htmd.simlist.Sim>` object
        The simulation.
    piece : int
        Which trajectory piece the frame corresponds to.
    frame : int
        The frame of the specific trajectory piece.
    """

    __slots__ = ["sim", "piece", "frame"]

    def __init__(self, sim, piece, frame):
        self.sim = sim
        self.piece = piece
        self.frame = frame

    def __repr__(self):
        return f"sim = {self.sim}\npiece = {self.piece}\nframe = {self.frame}"


class Sim(object):
    """Information on a single simulation trajectory.

    This class is used for storing information related to simulations. This information includes the path to the simulation,
    the path to a structure file (pdb) which corresponds to the simulation, the folder containing the input files used
    to generate the simulation (useful for adaptive), the parent of the simulation (if it was filtered it will point to
    the original simulation) and a unique simulation id.
    """

    def __init__(
        self,
        trajf,
        topology=None,
        inputf=None,
        _simid=None,
        _parent=None,
        _input=None,
        _trajectory=None,
        _molfile=None,
        _numframes=None,
    ):
        import uuid
        from htmd.util import ensurelist

        if trajf is not None:
            if isinstance(trajf, (list, tuple, np.ndarray)):
                _trajectory = [f for f in trajf]
            elif os.path.isdir(trajf):
                _trajectory = _autoDetectTrajectories(trajf)
            elif os.path.isfile(trajf):
                _trajectory = [trajf]
            else:
                raise ValueError(f"Invalid input structure: {trajf}")

            if topology is None:
                if os.path.isfile(trajf):
                    trajf = os.path.dirname(os.path.abspath(trajf))
                _molfile = _autoDetectTopology(trajf)
            else:
                if isinstance(topology, (list, tuple, np.ndarray)):
                    _molfile = [f for f in topology]
                elif os.path.isdir(topology):
                    _molfile = _autoDetectTopology(topology)
                else:
                    _molfile = [topology]

            if inputf is not None:
                _input = inputf
            else:
                _input = os.path.dirname(os.path.abspath(_molfile[0]))

            _numframes = [_readNumFrames(f) for f in _trajectory]
            if _simid is None:
                _simid = str(uuid.uuid4())

        self.simid = str(_simid)
        self.parent = _parent
        self.input = _input
        self.trajectory = _trajectory
        self.molfile = _molfile
        self.numframes = _numframes

    def toHDF5(self, h5group: h5py.Group):
        from moleculekit.util import ensurelist

        h5group.attrs["simid"] = self.simid
        h5group.create_dataset("input", data=ensurelist(self.input))
        h5group.create_dataset("trajectory", data=ensurelist(self.trajectory))
        h5group.create_dataset("molfile", data=ensurelist(self.molfile))
        h5group.create_dataset("numframes", data=ensurelist(self.numframes))
        if self.parent:
            parentgroup = h5group.create_group("parent")
            self.parent.toHDF5(parentgroup)

    @staticmethod
    def fromHDF5(h5group: h5py.Group):
        return Sim(
            trajf=None,
            _simid=str(h5group.attrs["simid"]),
            _input=[x.decode("utf-8") for x in h5group["input"]],
            _trajectory=[x.decode("utf-8") for x in h5group["trajectory"]],
            _molfile=[x.decode("utf-8") for x in h5group["molfile"]],
            _numframes=list(h5group["numframes"]),
            _parent=(
                None if "parent" not in h5group else Sim.fromHDF5(h5group["parent"])
            ),
        )

    def __repr__(self):
        parent = self.parent
        if self.parent is not None:
            parent = self.parent.simid
        return (
            f"\nsimid = {self.simid}\nparent = {parent}\ninput = {self.input}\ntrajectory = {self.trajectory}"
            f"\nmolfile = {self.molfile}\nnumframes = {self.numframes}\n"
        )

    def __eq__(self, other):
        iseq = True
        iseq &= self.simid == other.simid
        iseq &= np.all([x == y for x, y in zip(self.molfile, other.molfile)])
        iseq &= self.input == other.input
        iseq &= self.parent == other.parent
        iseq &= len(self.trajectory) == len(other.trajectory)
        if not iseq:
            return False
        for i in range(len(self.trajectory)):
            iseq &= self.trajectory[i] == other.trajectory[i]
        return iseq

    def copy(self):
        from copy import deepcopy

        return deepcopy(self)


def simlist(datafolders, topologies, inputfolders=None):
    """Creates a list of simulations

    Parameters
    ----------
    datafolders : str list
        A list of directories, each containing a single trajectory
    topologies : str list
        A list of topology files or folders containing a topology file corresponding to the trajectories in dataFolders.
        Can also be a single path to a single structure which corresponds to all trajectories.
        If the single path is a folder, topology files will be auto-detected in that folder.
    inputfolders : optional, str list
        A list of directories, each containing the input files used to produce the trajectories in dataFolders

    Return
    ------
    sims : np.ndarray of :class:`Sim <htmd.simlist.Sim>` objects
        A list of simulations

    Examples
    --------
    >>> simlist(glob('./test/data/*/'), './test/data/0/')
    >>> simlist(glob('./test/data/*/'), glob('./test/input/*/'), glob('./test/input/*/'))
    >>> simlist(glob('./test/data/*/'), glob('./test/input/*/*.pdb'), glob('./test/input/*/'))
    """
    from htmd.util import ensurelist
    import natsort

    if not datafolders:
        raise FileNotFoundError("No data folders were given, check your arguments.")
    if not topologies:
        raise FileNotFoundError("No molecule files were given, check your arguments.")
    topologies = ensurelist(topologies)
    datafolders = ensurelist(datafolders)
    for folder in datafolders:
        if not os.path.isdir(folder):
            raise NotADirectoryError(folder)
    if inputfolders:
        inputfolders = ensurelist(inputfolders)
        for folder in inputfolders:
            if not os.path.isdir(folder):
                raise NotADirectoryError(folder)

    # I need to match the simulation names inside the globs given. The
    # reason is that there can be more input folders in the glob than in
    # the data glob due to not having been retrieved. Hence I need to match
    # the folder names.

    # Create a hash map of data folder names
    datanames = dict()
    for folder in datafolders:
        if _simName(folder) in datanames:
            raise RuntimeError(
                "Duplicate simulation name detected. Cannot name-match directories."
            )
        datanames[_simName(folder)] = folder

    molnames = dict()
    for mol in topologies:
        if not os.path.exists(mol):
            raise FileNotFoundError(f"File {mol} does not exist")
        molnames[_simName(mol)] = mol

    if inputfolders:
        inputnames = dict()
        for inputf in inputfolders:
            inputnames[_simName(inputf)] = inputf

    logger.debug("Starting listing of simulations.")
    sims = []

    keys = natsort.natsorted(datanames.keys())
    i = 0
    from tqdm import tqdm

    for k in tqdm(keys, desc="Creating simlist"):
        trajectories = _autoDetectTrajectories(datanames[k])

        if not trajectories:
            continue

        if len(topologies) > 1:
            if k not in molnames:
                raise FileNotFoundError(
                    f"Did not find molfile with folder name {k} in the given glob"
                )
            molfile = molnames[k]
        else:
            molfile = topologies[0]

        if os.path.isdir(molfile):
            molfile = _autoDetectTopology(molfile)

        inputf = []
        if inputfolders:
            if k not in inputnames:
                raise FileNotFoundError(
                    f"Did not find input with folder name {k} in the given glob"
                )
            inputf = inputnames[k]

        numframes = [_readNumFrames(f) for f in trajectories]
        sims.append(
            Sim(
                trajf=None,
                _simid=i,
                _parent=None,
                _input=inputf,
                _trajectory=trajectories,
                _molfile=molfile,
                _numframes=numframes,
            )
        )
        i += 1
    logger.debug("Finished listing of simulations.")
    return sims


def simfilter(sims, outfolder, filtersel, njobs=None):
    """Filters a list of simulations generated by :func:`simlist`

    This function takes as input a list of simulations produced by `simList` and writes new trajectories containing only
    the desired atoms in a new directory.

    Parameters
    ----------
    sims : list
        A simulation list produced by the `simList` function
    outfolder : str
        The folder in which to write the modified trajectories
    filtersel : str
        Atom selection string describing the atoms we want to keep.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    njobs : int
        Number of parallel jobs to spawn for filtering of trajectories. If None it will use the default from htmd.config.

    Returns
    -------
    fsims : np.ndarray of :class:`Sim <htmd.simlist.Sim>` objects
        A list of filtered simulations

    Example
    -------
    >>> sims  = simlist(glob('data/*/'), glob('input/*/structure.pdb'))
    >>> fsims = simfilter(sims, 'filtered', filtersel='not water')
    """
    if not path.exists(outfolder):
        makedirs(outfolder)

    if len(sims) > 0:
        _filterTopology(sims[0], outfolder, filtersel)

    logger.debug("Starting filtering of simulations.")

    from htmd.config import _config
    from htmd.parallelprogress import ParallelExecutor, delayed

    aprun = ParallelExecutor(n_jobs=njobs if njobs is not None else _config["njobs"])
    filtsims = aprun(total=len(sims), desc="Filtering trajectories")(
        delayed(_filtSim)(i, sims, outfolder, filtersel) for i in range(len(sims))
    )

    logger.debug("Finished filtering of simulations")
    return np.array(filtsims)


def simmerge(simlist1, simlist2):
    """Merges two simlists by updating their `simid` fields

    Parameters
    ----------
    simlist1 : numpy array of :class:`Sim <htmd.simlist.Sim>` objects
        First list
    simlist2 : numpy array of :class:`Sim <htmd.simlist.Sim>` objects
        Second list

    Returns
    -------
    newlist : np.ndarray of :class:`Sim <htmd.simlist.Sim>` objects
        A new list containing all simulations
    """
    if len(simlist1) == 0:
        return simlist2
    if len(simlist2) == 0:
        return simlist1

    newsimlist = np.append(simlist1, simlist2)
    for i, s in enumerate(newsimlist):
        s.simid = i
    return newsimlist


def _filtSim(i, sims, outFolder, filterSel):
    name = _simName(sims[i].trajectory[0])
    directory = path.join(outFolder, name)
    if not path.exists(directory):
        makedirs(directory)

    logger.debug(f"Processing trajectory {name}")

    fmolfile = [
        path.join(outFolder, "filtered.psf"),
        path.join(outFolder, "filtered.pdb"),
    ]
    (traj, outtraj) = _renameSims(sims[i].trajectory, name, outFolder)
    if not traj:
        ftrajectory = _autoDetectTrajectories(path.join(outFolder, name))
        numframes = _getNumFrames(sims[i], ftrajectory)
        return Sim(
            trajf=None,
            _simid=sims[i].simid,
            _parent=sims[i],
            _input=None,
            _trajectory=ftrajectory,
            _molfile=fmolfile,
            _numframes=numframes,
        )

    try:
        from moleculekit.molecule import Molecule

        mol = Molecule(sims[i].molfile)
    except Exception as e:
        logger.warning(f"Error! Skipping simulation {name} due to error: {e}")
        return

    sel = mol.atomselect(filterSel)

    for j in range(0, len(traj)):
        try:
            mol.read(traj[j])
        except IOError as e:
            logger.warning(f"{e}, skipping trajectory")
            break

        mol.write(outtraj[j], sel)

    ftrajectory = _autoDetectTrajectories(path.join(outFolder, name))
    numframes = _getNumFrames(sims[i], ftrajectory)
    return Sim(
        trajf=None,
        _simid=sims[i].simid,
        _parent=sims[i],
        _input=None,
        _trajectory=ftrajectory,
        _molfile=fmolfile,
        _numframes=numframes,
    )


def _getNumFrames(sim, trajectories):
    if trajectories is None:
        return None
    numframes = sim.numframes
    if numframes is None or np.any([f is None for f in sim.numframes]):
        numframes = [_readNumFrames(f) for f in trajectories]
    return numframes


def _readNumFrames(filepath):
    filepath = os.path.abspath(filepath)
    filedir = os.path.dirname(filepath)
    basename = os.path.basename(filepath)
    numframefile = os.path.join(filedir, f".{basename}.numframes")
    numframes = None
    if os.path.exists(numframefile):
        with open(numframefile, "r") as f:
            try:
                numframes = int(f.readline())
            except Exception:
                raise RuntimeError(
                    f"{numframefile} does not contain an integer number of frames. Please delete this file."
                )
    return numframes


def _renameSims(trajectory, simname, outfolder):
    traj = list()
    outtraj = list()

    for t in range(0, len(trajectory)):
        (_, fname) = path.split(trajectory[t])
        (fname, ext) = path.splitext(fname)
        outname = path.join(outfolder, simname, f"{fname}.filtered{ext}")

        if not path.isfile(outname) or (
            path.getmtime(outname) < path.getmtime(trajectory[t])
        ):
            traj.append(trajectory[t])
            outtraj.append(outname)

    return traj, outtraj


def _filterTopology(sim, outfolder, filtsel):
    from htmd.util import ensurelist

    try:
        from moleculekit.molecule import Molecule

        mol = Molecule(sim.molfile)
    except IOError as e:
        raise RuntimeError(f"simFilter: {e}. Cannot read topology file {sim.molfile}")

    if (
        mol.coords.size == 0
    ):  # If we read for example psf or prmtop which have no coords, just add 0s everywhere
        mol.coords = np.zeros((mol.numAtoms, 3, 1), dtype=np.float32)

    extensions = [
        "pdb",
        "psf",
    ]  # Adding pdb and psf to make sure they are always written
    for m in ensurelist(sim.molfile):
        extensions.append(os.path.splitext(m)[1][1:])

    for ext in list(set(extensions)):
        filttopo = path.join(outfolder, f"filtered.{ext}")
        if not path.isfile(filttopo):
            try:
                mol.write(filttopo, filtsel)
            except Exception as e:
                logger.warning(
                    f"Filtering was not able to write {filttopo} due to error: {e}"
                )


def _autoDetectTrajectories(folder):
    from moleculekit.readers import _TRAJECTORY_READERS
    import natsort

    for tt in _TRAJECTORY_READERS:
        if tt in ("xsc",):
            # Some trajectory readers don't actually load trajectories (i.e. xsc files)
            continue
        trajectories = glob(path.join(folder, f"*.{tt}"))
        if len(trajectories) > 0:
            return natsort.natsorted(trajectories)


def _autoDetectTopology(folder):
    topo = {}
    for tt in ("prmtop", "psf", "mae", "pdb"):
        files = glob(path.join(folder, f"*.{tt}"))
        if len(files) > 0:
            if len(files) > 1:
                logger.warning(
                    f'Multiple "{tt}" files were found in folder {folder}. Picking {files[0]} as the topology'
                )
            topo[tt] = files[0]
            break

    if len(topo) == 0:
        raise RuntimeError(
            f"No topology file found in folder {folder}. "
            f"Supported extensions are {list(_TOPOLOGY_READERS.keys())}"
        )

    for tt in ("pdb", "coor"):
        files = glob(path.join(folder, f"*.{tt}"))
        if len(files) > 0:
            if len(files) > 1:
                logger.warning(
                    f'Multiple "{tt}" files were found in folder {folder}. Picking {files[0]} as the topology'
                )
            topo[tt] = files[0]
            break

    return list(topo.values())


def _simName(foldername):
    # Uses the name of the last folder as the simulation name
    if os.path.isdir(foldername):
        name = os.path.basename(os.path.normpath(foldername))
    else:
        name = os.path.basename(os.path.dirname(foldername))
    return name


class _TestSimlist(unittest.TestCase):
    def test_simlist_auto_structure(self):
        from htmd.home import home
        from htmd.projections.metric import _singleMolfile

        sims = simlist(
            glob(path.join(home(dataDir="adaptive"), "data", "*", "")),
            glob(path.join(home(dataDir="adaptive"), "input", "*")),
        )
        x = sims[0].copy()
        assert x == sims[0]
        assert x != sims[1]
        assert len(sims[0].molfile) == 2
        assert _singleMolfile(sims)[0]

    def test_simlist_many_structures(self):
        from htmd.home import home
        from htmd.projections.metric import _singleMolfile

        sims = simlist(
            glob(path.join(home(dataDir="adaptive"), "data", "*", "")),
            glob(path.join(home(dataDir="adaptive"), "input", "*", "structure.pdb")),
        )
        x = sims[0].copy()
        assert x == sims[0]
        assert x != sims[1]
        assert not isinstance(sims[0].molfile, list)
        assert _singleMolfile(sims)[0]

    def test_simlist_single_structure(self):
        from htmd.home import home
        from htmd.projections.metric import _singleMolfile

        sims = simlist(
            glob(path.join(home(dataDir="adaptive"), "data", "*", "")),
            path.join(home(dataDir="adaptive"), "input", "e1s1_1", "structure.pdb"),
        )
        x = sims[0].copy()
        assert x == sims[0]
        assert x != sims[1]
        assert not isinstance(sims[0].molfile, list)
        assert _singleMolfile(sims)[0]

    def test_autodetect(self):
        from htmd.home import home
        from htmd.projections.metric import _singleMolfile
        from natsort import natsorted

        datafolders = natsorted(
            glob(path.join(home(dataDir="adaptive"), "data", "*", ""))
        )
        inputfolders = natsorted(
            glob(path.join(home(dataDir="adaptive"), "input", "*"))
        )
        sims = []
        for ff in zip(datafolders, inputfolders):
            sims.append(Sim(ff[0], ff[1], ff[1]))

        x = sims[0].copy()
        assert x == sims[0]
        assert x != sims[1]
        assert len(sims[0].molfile) == 2
        assert _singleMolfile(sims)[0]

        sims2 = []
        for ff in zip(datafolders, inputfolders):
            sims2.append(Sim(ff[0], ff[1]))

        for s1, s2 in zip(sims, sims2):
            assert s1.parent == s2.parent
            assert s1.input == s2.input
            assert s1.trajectory == s2.trajectory
            assert s1.molfile == s2.molfile
            assert s1.numframes == s2.numframes

    def test_sim(self):
        import tempfile

        with tempfile.TemporaryDirectory() as tmpdir:
            with open(path.join(tmpdir, "test.pdb"), "w") as f:
                f.write("fake")
            with open(path.join(tmpdir, "test.psf"), "w") as f:
                f.write("fake")
            with open(path.join(tmpdir, "test.prmtop"), "w") as f:
                f.write("fake")
            with open(path.join(tmpdir, "test.mae"), "w") as f:
                f.write("fake")
            with open(path.join(tmpdir, "test.coor"), "w") as f:
                f.write("fake")
            for i in range(11):
                with open(path.join(tmpdir, f"test_{i}.xtc"), "w") as f:
                    f.write("fake")

            # Passing a single directory
            sim = Sim(tmpdir)
            assert sim.molfile == [
                path.join(tmpdir, "test.prmtop"),
                path.join(tmpdir, "test.pdb"),
            ]
            assert sim.trajectory == [
                path.join(tmpdir, f"test_{i}.xtc") for i in range(11)
            ]
            assert sim.input == tmpdir

            # Passing same directory to all arguments
            sim = Sim(tmpdir, tmpdir, tmpdir)
            assert sim.molfile == [
                path.join(tmpdir, "test.prmtop"),
                path.join(tmpdir, "test.pdb"),
            ]
            assert sim.trajectory == [
                path.join(tmpdir, f"test_{i}.xtc") for i in range(11)
            ]
            assert sim.input == tmpdir

            # Passing single topology file and arbitrary input folder
            sim = Sim(tmpdir, path.join(tmpdir, "test.pdb"), "/tmp")
            assert sim.molfile == [path.join(tmpdir, "test.pdb")]
            assert sim.trajectory == [
                path.join(tmpdir, f"test_{i}.xtc") for i in range(11)
            ]
            assert sim.input == "/tmp"

        # Coor file reading
        with tempfile.TemporaryDirectory() as tmpdir:
            with open(path.join(tmpdir, "test.psf"), "w") as f:
                f.write("fake")
            with open(path.join(tmpdir, "test.mae"), "w") as f:
                f.write("fake")
            with open(path.join(tmpdir, "test.coor"), "w") as f:
                f.write("fake")
            for i in range(11):
                with open(path.join(tmpdir, f"test_{i}.xtc"), "w") as f:
                    f.write("fake")

            sim = Sim(tmpdir)
            assert sim.molfile == [
                path.join(tmpdir, "test.psf"),
                path.join(tmpdir, "test.coor"),
            ]
            assert sim.trajectory == [
                path.join(tmpdir, f"test_{i}.xtc") for i in range(11)
            ]
            assert sim.input == tmpdir


if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from __future__ import print_function

import numpy as np
import os.path as path
import os
import re
import shutil
import textwrap

from subprocess import call
from htmd.home import home
from moleculekit.molecule import Molecule
from moleculekit.util import _missingSegID
from htmd.builder.builder import (
    detectDisulfideBonds,
    convertDisulfide,
    detectCisPeptideBonds,
    _checkMixedSegment,
    _checkLongResnames,
    MissingResidueError,
    BuildError,
)
from htmd.builder.ionize import ionize as ionizef, ionizePlace
from glob import glob
import unittest


import logging

logger = logging.getLogger(__name__)

_psfgen_exists = shutil.which("psfgen", mode=os.X_OK) is not None


def htmdCharmmHome():
    """Returns the location of the CHARMM files distributed with HTMD"""
    return os.path.abspath(
        os.path.join(home(shareDir=True), "builder", "charmmfiles", "")
    )


def listFiles():
    """Lists all available Charmm topologies and parameter files

    Examples
    --------
    >>> from htmd.builder import charmm
    >>> charmm.listFiles()             # doctest: +ELLIPSIS
    ---- Topologies files list...

    """
    from natsort import natsorted

    charmmdir = htmdCharmmHome()
    topos = natsorted(glob(path.join(charmmdir, "top", "*.rtf")))
    params = natsorted(glob(path.join(charmmdir, "par", "*.prm")))
    streams = natsorted(glob(path.join(charmmdir, "str", "*", "*.str")))
    print("---- Topologies files list: " + path.join(charmmdir, "top", "") + " ----")
    for t in topos:
        t = t.replace(charmmdir, "")
        print(t)
    print("---- Parameters files list: " + path.join(charmmdir, "par", "") + " ----")
    for p in params:
        p = p.replace(charmmdir, "")
        print(p)
    print("---- Stream files list: " + path.join(charmmdir, "str", "") + " ----")
    for s in streams:
        s = s.replace(charmmdir, "")
        print(s)


def search(key, name):
    """Searches for CHARMM files containing a given definition.

    Parameters
    ----------
    key : str
        A key
    name : str
        The corresponding name

    Examples
    --------
    >>> charmm.search(key='RESI', name = 'CHL1')  # doctest: +SKIP
    """
    os.system(
        'find {} -type f -exec grep -n "{} {}" {{}} +'.format(
            htmdCharmmHome(), key, name
        )
    )


def defaultTopo():
    """Returns the default topologies used by charmm.build"""
    return [
        "top/top_all36_prot.rtf",
        "top/top_all36_lipid.rtf",
        "top/top_water_ions.rtf",
        "top/top_all36_cgenff.rtf",
    ]


def defaultParam():
    """Returns the default parameters used by charmm.build"""
    return [
        "par/par_all36m_prot.prm",
        "par/par_all36_lipid.prm",
        "par/par_water_ions.prm",
        "par/par_all36_cgenff.prm",
    ]


def defaultStream():
    """Returns the default stream files used by charmm.build"""
    return ["str/prot/toppar_all36_prot_arg0.str", "str/misc/toppar_ions_won.str"]


def build(
    mol,
    topo=None,
    param=None,
    stream=None,
    prefix="structure",
    outdir="./build",
    caps=None,
    ionize=True,
    saltconc=0,
    saltanion=None,
    saltcation=None,
    disulfide=None,
    regenerate=["angles", "dihedrals"],
    patches=None,
    noregen=None,
    aliasresidues=None,
    psfgen=None,
    execute=True,
    _clean=True,
):
    """Builds a system for CHARMM

    Uses VMD and psfgen to build a system for CHARMM. Additionally it allows for ionization and adding of disulfide bridges.

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object containing the system
    topo : list of str
        A list of topology `rtf` files.
        Use :func:`charmm.listFiles <htmd.builder.charmm.listFiles>` to get a list of available topology files.
        Default: ['top/top_all36_prot.rtf', 'top/top_all36_lipid.rtf', 'top/top_water_ions.rtf']
    param : list of str
        A list of parameter `prm` files.
        Use :func:`charmm.listFiles <htmd.builder.charmm.listFiles>` to get a list of available parameter files.
        Default: ['par/par_all36m_prot.prm', 'par/par_all36_lipid.prm', 'par/par_water_ions.prm']
    stream : list of str
        A list of stream `str` files containing topologies and parameters.
        Use :func:`charmm.listFiles <htmd.builder.charmm.listFiles>` to get a list of available stream files.
        Default: ['str/prot/toppar_all36_prot_arg0.str']
    prefix : str
        The prefix for the generated pdb and psf files
    outdir : str
        The path to the output directory
        Default: './build'
    caps : dict
        A dictionary with keys segids and values lists of strings describing the caps of that segment.
        e.g. caps['P'] = ['first ACE', 'last CT3'] or caps['P'] = ['first none', 'last none'].
        Default: will apply ACE and CT3 caps to proteins and none caps to the rest.
    ionize : bool
        Enable or disable ionization
    saltconc : float
        Salt concentration (in Molar) to add to the system after neutralization.
    saltanion : {'CLA'}
        The anion type. Please use only CHARMM ion atom names.
    saltcation : {'SOD', 'MG', 'POT', 'CES', 'CAL', 'ZN2'}
        The cation type. Please use only CHARMM ion atom names.
    disulfide : list of pairs of atomselection strings
        If None it will guess disulfide bonds. Otherwise provide a list pairs of atomselection strings for each pair of
        residues forming the disulfide bridge.
    regenerate : None or list of strings of: ['angles', 'dihedrals']
        Disable angle/dihedral regeneration with `regenerate=None`, or enable it with `regenerate=['angles', 'diheldrals']`
        or just one of the two options with `regenerate=['angles']` or `regenerate=['diheldrals']`.
    patches : list of str
        Any further patches the user wants to apply
    noregen : list of str
        A list of patches that must not be regenerated (angles and dihedrals)
        Default: ['FHEM', 'PHEM', 'PLOH', 'PLO2', 'PLIG', 'PSUL']
    aliasresidues : dict of aliases
        A dictionary of key: value pairs of residue names we want to alias
    psfgen : str
        Path to psfgen executable used to build for CHARMM
    execute : bool
        Disable building. Will only write out the input script needed by psfgen. Does not include ionization.

    Returns
    -------
    molbuilt : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The built system in a Molecule object

    Example
    -------
    >>> from htmd.ui import *
    >>> mol = Molecule("3PTB")
    >>> mol.filter("not resname BEN")
    >>> molbuilt = charmm.build(mol, outdir='/tmp/build', ionize=False)  # doctest: +ELLIPSIS
    Bond between A: [serial 185 resid 42 resname CYS chain A segid 0]
                 B: [serial 298 resid 58 resname CYS chain A segid 0]...
    >>> # More complex example
    >>> topos  = ['top/top_all36_prot.rtf', './BEN.rtf', 'top/top_water_ions.rtf']
    >>> params = ['par/par_all36m_prot.prm', './BEN.prm', 'par/par_water_ions.prm']
    >>> disu = [['segid P and resid 157', 'segid P and resid 13'], ['segid K and resid 1', 'segid K and resid 25']]
    >>> ar = {'SAPI24': 'SP24'}  # Alias large resnames to a short-hand version
    >>> molbuilt = charmm.build(mol, topo=topos, param=params, outdir='/tmp/build', saltconc=0.15, disulfide=disu, aliasresidues=ar)  # doctest: +SKIP
    """

    mol = mol.copy()
    _missingSegID(mol)
    _checkMixedSegment(mol)
    _checkLongResnames(mol, aliasresidues)
    if psfgen is None:
        psfgen = shutil.which("psfgen", mode=os.X_OK)
        if not psfgen:
            raise FileNotFoundError(
                "Could not find psfgen executable, or no execute permissions are given. "
                "Run `conda install psfgen -c acellera`."
            )
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    if _clean:
        _cleanOutDir(outdir)
    if topo is None:
        topo = defaultTopo()
    if param is None:
        param = defaultParam()
    if stream is None:
        stream = defaultStream()
    if caps is None:
        caps = _defaultCaps(mol)
    # patches that must _not_ be regenerated
    if noregen is None:
        noregen = ["FHEM", "PHEM", "PLOH", "PLO2", "PLIG", "PSUL"]

    alltopo = topo.copy()
    allparam = param.copy()

    # Splitting the stream files and adding them to the list of parameter and topology files
    charmmdir = htmdCharmmHome()
    for s in stream:
        if s[0] != "." and path.isfile(path.join(charmmdir, s)):
            s = path.join(charmmdir, s)
        outrtf, outprm = _prepareStream(s)
        alltopo.append(outrtf)
        allparam.append(outprm)

    # _missingChain(mol)
    # _checkProteinGaps(mol)
    if patches is None:
        patches = []
    if isinstance(patches, str):
        patches = [patches]
    allpatches = []
    allpatches += patches
    # Find protonated residues and add patches for them
    allpatches += _protonationPatches(mol)

    f = open(path.join(outdir, "build.vmd"), "w")
    f.write("# psfgen file generated by charmm.build\n")
    f.write("package require psfgen;\n")
    f.write("psfcontext reset;\n\n")

    # Copying and printing out the topologies
    if not path.exists(path.join(outdir, "topologies")):
        os.makedirs(path.join(outdir, "topologies"))
    for i in range(len(alltopo)):
        if alltopo[i][0] != "." and path.isfile(path.join(charmmdir, alltopo[i])):
            alltopo[i] = path.join(charmmdir, alltopo[i])
        localname = "{}.".format(i) + path.basename(alltopo[i])
        shutil.copy(alltopo[i], path.join(outdir, "topologies", localname))
        f.write("topology " + path.join("topologies", localname) + "\n")
    f.write("\n")

    _printAliases(f)
    if aliasresidues is not None:  # User defined aliases
        for key, val in aliasresidues.items():
            mol.resname[mol.resname == key] = val
            f.write("        pdbalias residue {} {}\n".format(val, key))

    # Printing out segments
    if not path.exists(path.join(outdir, "segments")):
        os.makedirs(path.join(outdir, "segments"))
    logger.info("Writing out segments.")
    segments = _getSegments(mol)
    wateratoms = mol.atomselect("water")
    for seg in segments:
        pdbname = "segment" + seg + ".pdb"
        segatoms = mol.segid == seg
        mol.write(path.join(outdir, "segments", pdbname), sel=segatoms)

        segwater = wateratoms & segatoms
        f.write("segment " + seg + " {\n")
        if np.all(
            segatoms == segwater
        ):  # If segment only contains waters, set: auto none
            f.write("\tauto none\n")
        f.write("\tpdb " + path.join("segments", pdbname) + "\n")
        if caps is not None and seg in caps:
            for c in caps[seg]:
                f.write("\t" + c + "\n")
        f.write("}\n")
        f.write("coordpdb " + path.join("segments", pdbname) + " " + seg + "\n\n")

    if (
        disulfide is not None
        and len(disulfide) != 0
        and isinstance(disulfide[0][0], str)
    ):
        disulfide = convertDisulfide(mol, disulfide)

    if disulfide is None:
        disulfide = detectDisulfideBonds(mol)

    if len(disulfide) != 0:
        for d in sorted(disulfide, key=lambda x: x[0].segid):
            str0 = f"{d[0].segid}:{d[0].resid}{d[0].insertion}"
            str1 = f"{d[1].segid}:{d[1].resid}{d[1].insertion}"
            f.write(f"patch DISU {str0} {str1}\n")
        f.write("\n")

    noregenpatches = [p for p in allpatches if p.split()[1] in noregen]
    regenpatches = [p for p in allpatches if p.split()[1] not in noregen]

    # Printing regenerable patches
    if len(regenpatches) != 0:
        for p in regenpatches:
            f.write(p + "\n")
        f.write("\n")

    # Regenerate angles and dihedrals
    if regenerate is not None:
        f.write("regenerate {}\n".format(" ".join(regenerate)))
        f.write("\n")

    # Printing non-regenerable patches
    if len(noregenpatches) != 0:
        for p in noregenpatches:
            f.write(p + "\n")
        f.write("\n")

    f.write("guesscoord\n")
    f.write("writepsf " + prefix + ".psf\n")
    f.write("writepdb " + prefix + ".pdb\n")
    # f.write('quit\n')
    f.close()

    if allparam is not None:
        combine(allparam, path.join(outdir, "parameters"))

    molbuilt = None
    if execute:
        logpath = os.path.abspath("{}/log.txt".format(outdir))
        logger.info("Starting the build.")
        currdir = os.getcwd()
        os.chdir(outdir)
        f = open(logpath, "w")
        # call([vmd, '-dispdev', 'text', '-e', './build.vmd'], stdout=f)
        my_env = os.environ.copy()
        my_env["LC_ALL"] = "C"
        call([psfgen, "./build.vmd"], stdout=f, stderr=f, env=my_env)
        f.close()
        errors = _logParser(logpath)
        os.chdir(currdir)
        if errors:
            raise BuildError(
                errors
                + [
                    "Check {} for further information on errors in building.".format(
                        logpath
                    )
                ],
                errors,
            )
        logger.info("Finished building.")

        if path.isfile(path.join(outdir, "structure.pdb")) and path.isfile(
            path.join(outdir, "structure.psf")
        ):
            molbuilt = Molecule(path.join(outdir, "structure.pdb"))
            molbuilt.read(path.join(outdir, "structure.psf"))
        else:
            raise BuildError(
                "No structure pdb/psf file was generated. Check {} for errors in building.".format(
                    logpath
                )
            )

        if ionize:
            os.makedirs(path.join(outdir, "pre-ionize"))
            data = glob(path.join(outdir, "*"))
            for f in data:
                shutil.move(f, path.join(outdir, "pre-ionize"))
            totalcharge = np.sum(molbuilt.charge)
            nwater = np.sum(molbuilt.atomselect("water and noh"))
            anion, cation, anionatom, cationatom, nanion, ncation = ionizef(
                molbuilt,
                totalcharge,
                nwater,
                saltconc=saltconc,
                anion=saltanion,
                cation=saltcation,
            )
            water_sel = mol.atomselect("water")
            solvent_mol = mol.copy(sel=water_sel)
            solute_mol = mol.copy(sel=~water_sel)
            solvent_mol = ionizePlace(
                solvent_mol,
                solute_mol,
                anion,
                cation,
                anionatom,
                cationatom,
                nanion,
                ncation,
            )
            newmol = solute_mol.copy()
            newmol.append(solvent_mol)
            # Redo the whole build but now with ions included
            return build(
                newmol,
                topo=alltopo,
                param=allparam,
                stream=[],
                prefix=prefix,
                outdir=outdir,
                ionize=False,
                caps=caps,
                execute=execute,
                saltconc=saltconc,
                disulfide=disulfide,
                regenerate=regenerate,
                patches=patches,
                noregen=noregen,
                aliasresidues=aliasresidues,
                psfgen=psfgen,
                _clean=False,
            )
    _checkFailedAtoms(molbuilt)
    _recoverProtonations(molbuilt)
    detectCisPeptideBonds(molbuilt, respect_bonds=True)  # Warn in case of cis bonds
    cleanup_charmm_prm(
        molbuilt, path.join(outdir, "parameters"), path.join(outdir, "parameters.prm")
    )
    return molbuilt


def _cleanOutDir(outdir):
    from glob import glob

    files = glob(os.path.join(outdir, "structure.*"))
    files += glob(os.path.join(outdir, "log.*"))
    files += glob(os.path.join(outdir, "*.log"))
    files += glob(os.path.join(outdir, "*.vmd"))
    files += glob(os.path.join(outdir, "parameters"))
    for f in files:
        os.remove(f)
    folders = glob(os.path.join(outdir, "segments"))
    folders += glob(os.path.join(outdir, "topologies"))
    folders += glob(os.path.join(outdir, "pre-ionize"))
    for f in folders:
        shutil.rmtree(f)


def _getSegments(mol):
    # Calculate unique segments but keep sorting
    indexes = np.unique(mol.segid, return_index=True)[1]
    uqseg = [mol.segid[index] for index in sorted(indexes)]
    return uqseg


def _printAliases(f):
    lines = """
        # Aliases
        pdbalias residue G GUA
        pdbalias residue C CYT
        pdbalias residue A ADE
        pdbalias residue T THY
        pdbalias residue U URA

        foreach bp { GUA CYT ADE THY URA } {
            pdbalias atom $bp "O5\*" O5'
            pdbalias atom $bp "C5\*" C5'
            pdbalias atom $bp "O4\*" O4'
            pdbalias atom $bp "C4\*" C4'
            pdbalias atom $bp "C3\*" C3'
            pdbalias atom $bp "O3\*" O3'
            pdbalias atom $bp "C2\*" C2'
            pdbalias atom $bp "O2\*" O2'
            pdbalias atom $bp "C1\*" C1'
        }

        pdbalias atom ILE CD1 CD
        pdbalias atom SER HG HG1
        pdbalias residue HIS HSD

        # Heme aliases
        pdbalias residue HEM HEME
        pdbalias atom HEME "N A" NA
        pdbalias atom HEME "N B" NB
        pdbalias atom HEME "N C" NC
        pdbalias atom HEME "N D" ND

        # Water aliases
        pdbalias residue HOH TIP3
        pdbalias atom TIP3 O OH2

        # Ion aliases
        pdbalias residue K POT
        pdbalias atom K K POT
        pdbalias residue ICL CLA
        pdbalias atom ICL CL CLA
        pdbalias residue INA SOD
        pdbalias atom INA NA SOD
        pdbalias residue CA CAL
        pdbalias atom CA CA CAL
        pdbalias residue ZN ZN2
        pdbalias residue FE2 Fe2p
        pdbalias atom Fe2p FE Fe2p 

        # Other aliases
        pdbalias atom LYS 1HZ HZ1
        pdbalias atom LYS 2HZ HZ2
        pdbalias atom LYS 3HZ HZ3

        pdbalias atom ARG 1HH1 HH11
        pdbalias atom ARG 2HH1 HH12
        pdbalias atom ARG 1HH2 HH21
        pdbalias atom ARG 2HH2 HH22

        pdbalias atom ASN 1HD2 HD21
        pdbalias atom ASN 2HD2 HD22

        # Aliases for PDB ions / elements
        pdbalias residue LI LIT
        pdbalias residue NA SOD
        pdbalias residue K POT
        pdbalias residue CA CAL
        pdbalias residue ZN ZN2
        pdbalias residue CL CLA
        pdbalias residue RB RUB
        pdbalias residue CD CD2
        pdbalias residue CS CES
        pdbalias residue BA BAR
        pdbalias residue NI Ni2p
        pdbalias residue HG Hg2p

        # Aliases for Maestro residues
        pdbalias residue AR0 ARG
        pdbalias residue GLH GLU
        pdbalias residue ASH ASP
        pdbalias residue LYN LYS
        pdbalias residue HIE HSE
        pdbalias residue HID HSD
        pdbalias residue HIP HSP
        pdbalias residue CYX CYS
        pdbalias residue CYM CYS
        pdbalias residue WAT TIP3

        # Generated by gen_psfaliases.py (Toni) on 2016-03-11 10:21
        pdbalias atom ALA H HN
        pdbalias atom ARG H HN
        pdbalias atom ARG HB3 HB1
        pdbalias atom ARG HG3 HG1
        pdbalias atom ARG HD3 HD1
        pdbalias atom ASP H HN
        pdbalias atom ASP HB3 HB1
        pdbalias atom ASN H HN
        pdbalias atom ASN HB3 HB1
        pdbalias atom CYS H HN
        pdbalias atom CYS HB3 HB1
        pdbalias atom GLU H HN
        pdbalias atom GLU HB3 HB1
        pdbalias atom GLU HG3 HG1
        pdbalias atom GLN H HN
        pdbalias atom GLN HB3 HB1
        pdbalias atom GLN HG3 HG1
        pdbalias atom GLY H HN
        pdbalias atom GLY HA3 HA1
        pdbalias atom HIS H HN
        pdbalias atom HIS HB3 HB1
        pdbalias atom ILE H HN
        pdbalias atom ILE HG13 HG11
        pdbalias atom ILE HD11 HD1
        pdbalias atom ILE HD12 HD2
        pdbalias atom ILE HD13 HD3
        pdbalias atom LEU H HN
        pdbalias atom LEU HB3 HB1
        pdbalias atom LYS H HN
        pdbalias atom LYS HB3 HB1
        pdbalias atom LYS HG3 HG1
        pdbalias atom LYS HD3 HD1
        pdbalias atom LYS HE3 HE1
        pdbalias atom MET H HN
        pdbalias atom MET HB3 HB1
        pdbalias atom MET HG3 HG1
        pdbalias atom PHE H HN
        pdbalias atom PHE HB3 HB1
        pdbalias atom PRO H2 HT2
        pdbalias atom PRO H3 HT1
        pdbalias atom PRO HB3 HB1
        pdbalias atom PRO HG3 HG1
        pdbalias atom PRO HD3 HD1
        pdbalias atom SER H HN
        pdbalias atom SER HB3 HB1
        pdbalias atom THR H HN
        pdbalias atom TRP H HN
        pdbalias atom TRP HB3 HB1
        pdbalias atom TYR H HN
        pdbalias atom TYR HB3 HB1
        pdbalias atom VAL H HN

        # Aliases for carbohydrates
        pdbalias residue NAG BGLCNA
        pdbalias residue BMA BMAN
        pdbalias residue GLA AGAL
        pdbalias residue GAL BGAL
        pdbalias residue MAN AMAN
        pdbalias residue FUC AFUC
        pdbalias residue FUL BFUC

    """
    f.write(textwrap.dedent(lines))
    f.write("\n\n")


def _defaultCaps(mol):
    # neutral for protein, nothing for any other segment
    # of course this might not be ideal for protein which require charged terminals
    prot = mol.atomselect("protein")
    segsProt = np.unique(mol.segid[prot])
    segsNonProt = np.unique(mol.segid[~prot])
    caps = dict()
    for s in segsProt:
        if len(np.unique(mol.resid[mol.segid == s])) < 10:
            logger.warning(
                "Segment {} consists of a peptide with less than 10 residues. It will not be capped by "
                "default. If you want to cap it use the caps argument of charmm.build to manually define "
                "caps for all segments".format(s)
            )
            continue
        nter, cter = _removeCappedResidues(mol, s)
        caps[s] = ["first {}".format(nter), "last {}".format(cter)]
    for s in segsNonProt:
        caps[s] = ["first none", "last none"]
    return caps


def _removeCappedResidues(mol, seg):
    # Default caps for charmm
    nter = "ACE"
    cter = "CT3"

    # Mapping from various residue caps to charmm patches
    """
    CHARMM patches:
    # N-terminus patches
    NTER         1.00 ! standard N-terminus
    GLYP         1.00 ! Glycine N-terminus
    PROP         1.00 ! Proline N-Terminal
    ACE          0.00 ! acetylated N-terminus
    ACED         0.00 ! acetylated N-terminus (to create dipeptide)
    ACP          0.00 ! acetylated N-terminus for proline
    ACPD         0.00 ! acetylated N-terminus for proline (to create dipeptide)
    NNEU         0.00 ! neutral N-terminus; charges from LSN
    # C-Terminus patches
    CTER        -1.00 ! standard C-terminus
    CNEU         0.00 ! protonated (neutral) C-terminu, charges from ASPP
    CTP          0.00 ! protonated C-terminus
    CT1          0.00 ! methylated C-terminus from methyl acetate
    CT2          0.00 ! amidated C-terminus
    CT3          0.00 ! N-Methylamide C-terminus
    """

    ntercaps = {"ACE": "ACE"}
    ctercaps = {"NMA": "CT3"}

    # If caps already exist, remove them and convert them to charmm caps
    for n in ntercaps:
        asel = (mol.segid == seg) & (mol.resname == n)
        if np.sum(asel) != 0:
            mol.remove(asel, _logger=False)
            nter = ntercaps[n]
            break  # I expect only one capped n-terminal residue in one segment!
    for c in ctercaps:
        asel = (mol.segid == seg) & (mol.resname == c)
        if np.sum(asel) != 0:
            mol.remove(asel, _logger=False)
            cter = ctercaps[c]
            break  # I expect only one capped c-terminal residue in one segment!
    return nter, cter


# Mapping Maestro protonated residue names to CHARMM patches
def _protonationPatches(mol):
    protonations = {
        "GLH": "GLUP",
        "ASH": "ASPP",
        "LYN": "LSN",
        "AR0": "RN1",
        "CYM": "CYSD",
    }
    aliases = {}  # Some protonations don't exist in CHARMM
    # TODO: Remember to alias all of them before applying patches
    patches = []

    for pro in sorted(protonations.keys()):
        sel = (mol.resname == pro) & (mol.name == "CA")
        pseg = mol.segid[sel]
        pres = mol.resid[sel]
        if len(pseg) == 0:
            continue
        for r in range(len(pseg)):
            patch = "patch {} {}:{}".format(protonations[pro], pseg[r], pres[r])
            patches.append(patch)

    """for pro in aliases:
        sel = mol.atomselect('resname {}'.format(pro))
        if np.sum(sel) != 0:
            logger.warning('Found resname {}. This protonation state does not exist in CHARMM '
                           'and will be reverted to {}.'.format(pro, aliases[pro]))
            mol.set('resname', aliases[pro], sel=sel)"""
    for pro in aliases:
        sel = mol.resname == pro
        if np.any(sel):
            raise RuntimeError(
                "Found resname {}. This protonation state does not exist in CHARMM. Cannot build."
            )

    return patches


# Recover protonation states of residues after building (CHARMM renames protonated residues to standard names)
def _recoverProtonations(mol):
    mol.set("resname", "ASH", sel="same residue as (resname ASP and name HD2)")
    mol.set("resname", "GLH", sel="same residue as (resname GLU and name HE2)")
    mol.set(
        "resname",
        "LYN",
        sel="resname LYS and not (same residue as (resname LYS and name HZ3))",
    )  # The LYN patch removes the HZ3 proton
    mol.set(
        "resname",
        "AR0",
        sel="resname ARG and not (same residue as (resname ARG and name HE))",
    )  # The AR0 patch removes the HE proton
    # Histidine protonations keep their names in CHARMM. No need to rename them


def combine(prmlist, outfile):
    """Combines CHARMM parameter files
    Take a list of parameters files and combine them into a single file (useful for acemd)

    Parameters
    ----------
    prmlist: list
        List of parameter files to combine
    outfile: str
        Output filename of combined parameter files

    """
    # Process parameter files
    prm_list = [
        "!COMMENTS\n",
        "ATOMS\n",
        "BONDS\n",
        "ANGLES\n",
        "DIHEDRALS\n",
        "IMPROPER\n",
        "CMAP\n",
        "NONBONDED\n",
        "NBFIX\n",
        "HBOND\n",
    ]

    charmmdir = htmdCharmmHome()
    for myfile in prmlist:
        if myfile[0] != "." and path.isfile(path.join(charmmdir, myfile)):
            myfile = path.join(charmmdir, myfile)
        if not path.isfile(myfile):
            raise FileNotFoundError(
                myfile + " file does not exist. Cannot create combined parameter file."
            )
        fn = os.path.basename(myfile)
        with open(myfile, "r", encoding="utf-8") as fh:
            context = 0
            for line in fh:
                if re.search(r"^ATOMS", line):
                    context = 1
                    prm_list[context] += _sec_name(fn)
                elif re.search(r"^BOND", line):
                    context = 2
                    prm_list[context] += _sec_name(fn)
                elif re.search(r"^ANGL", line):
                    context = 3
                    prm_list[context] += _sec_name(fn)
                elif re.search(r"^DIHE", line) or re.search(r"^THET", line):
                    context = 4
                    prm_list[context] += _sec_name(fn)
                elif re.search(r"^IMPR", line) or re.search(r"^IMPH", line):
                    context = 5
                    prm_list[context] += _sec_name(fn)
                elif re.search(r"^CMAP", line) or re.search(r"^NBON", line):
                    context = 6
                    prm_list[context] += _sec_name(fn)
                elif re.search(r"^NONB", line):
                    context = 7
                    prm_list[context] += _sec_name(fn)
                elif re.search(r"^NBFI", line):
                    context = 8
                    prm_list[context] += _sec_name(fn)
                elif re.search(r"^HBON", line):
                    context = 9
                    prm_list[context] += _sec_name(fn)
                else:
                    if context == 0:  # COMMENTS
                        if (
                            re.search(r"^\s*\!+", line, re.I)
                            or re.search(r"^\s*\*+", line, re.I)
                            or len(line.strip()) == 0
                        ):
                            prm_list[context] += line
                        else:
                            prm_list[context] += "!" + line
                    elif not line.lower().startswith(
                        "end"
                    ) and not line.lower().startswith("return"):
                        prm_list[context] += line

    prm = "".join(map(str, prm_list)) + "END"
    prmfh = open(outfile, "w")
    prmfh.write(prm)
    prmfh.close()


# Create string to indicate source of section
def _sec_name(filename):
    return "!Following lines added from %s\n" % (filename)


def split(filename, outdir):
    """Splits a stream file into an rtf and prm file.

    Parameters
    ----------
    filename : str
        Stream file name
    """
    regex = re.compile(r"^(toppar_)?(.*)\.str$")
    base = os.path.basename(os.path.normpath(filename))
    base = regex.findall(base)[0][1]
    outrtf = os.path.join(outdir, f"top_{base}.rtf")
    outprm = os.path.join(outdir, f"par_{base}.prm")

    startrtf = re.compile(r"^read rtf card", flags=re.IGNORECASE)
    startprm = re.compile(r"^read para\w* card", flags=re.IGNORECASE)
    endsection = re.compile(r"^end", flags=re.IGNORECASE)

    rtfsection = 0
    prmsection = 0
    section = "junk"

    rtfstr = ""
    prmstr = ""

    f = open(filename, "r")
    for line in f:
        if startrtf.match(line):
            rtfsection += 1
            if rtfsection > 1:
                rtfstr += "! WARNING -- ANOTHER rtf SECTION FOUND\n"
            section = "rtf"
        elif startprm.match(line):
            prmsection += 1
            if prmsection > 1:
                prmstr += "! WARNING -- ANOTHER para SECTION FOUND\n"
            section = "prm"
        elif endsection.match(line):
            section = "junk"
        elif section == "rtf":
            rtfstr += line
        elif section == "prm":
            prmstr += line
    f.close()

    if rtfsection > 1:
        raise BuildError(
            "Multiple ({}) rtf topology sections found in {} stream file.".format(
                rtfsection, filename
            )
        )
    if prmsection > 1:
        raise BuildError(
            "Multiple ({}) prm parameter sections found in {} stream file.".format(
                prmsection, filename
            )
        )

    f = open(outrtf, "w")
    f.write(rtfstr + "END\n")
    f.close()
    f = open(outprm, "w")
    f.write(prmstr + "END\n")
    f.close()
    return outrtf, outprm


def _prepareStream(filename):
    from htmd.util import tempname

    tmpout = tempname()
    os.makedirs(tmpout)
    return split(filename, tmpout)


def _logParser(fname):
    import re

    unknownres_regex = re.compile(r"unknown residue type (\w+)")
    failedcoor = re.compile(r"Warning: failed to set coordinate for atom")
    failedangle = re.compile(r"Warning: failed to guess coordinate due to bad angle")
    poorlycoor = re.compile(r"Warning: poorly guessed coordinate(s?)")
    otherwarn = re.compile(r"Warning")

    failedcoorcount = 0
    failedanglecount = 0
    poorlycoorcount = -1  # Discount the summary report message in the log
    otherwarncount = 0
    unknownres = []
    with open(fname, "r") as f:
        for line in f:
            if failedcoor.search(line):
                failedcoorcount += 1
            elif failedangle.search(line):
                failedanglecount += 1
            elif poorlycoor.search(line):
                poorlycoorcount += 1
            elif otherwarn.search(line):
                otherwarncount += 1
            elif unknownres_regex.search(line):
                unknownres.append(unknownres_regex.findall(line)[0])

    warnings = False
    errors = []
    if failedcoorcount > 0:
        warnings = True
        logger.warning(f"Failed to set coordinates for {failedcoorcount} atoms.")
    if failedanglecount > 0:
        warnings = True
        logger.warning(
            f"Failed to guess coordinates for {failedanglecount} atoms due to bad angles."
        )
    if poorlycoorcount > 0:
        warnings = True
        logger.warning(f"Poorly guessed coordinates for {poorlycoorcount} atoms.")
    if otherwarncount > 0:
        warnings = True
        logger.warning(
            f"{otherwarncount} undefined warnings were produced during building."
        )
    if len(unknownres):
        errors.append(
            MissingResidueError(
                f"Unknown residue(s) {np.unique(unknownres)} found in the input structure. "
                "You are either missing a topology definition for the residue or you need to "
                "rename it to the correct residue name",
                unknownres,
            )
        )
    if warnings:
        logger.warning(f"Please check {fname} for further information.")

    return errors


def _checkFailedAtoms(mol):
    if mol is None:
        return
    idx = np.where(np.sum(mol.coords == 0, axis=1) == 3)[0]
    if len(idx) != 0:
        logger.critical(
            f"Atoms with indexes {idx} are positioned at [0,0,0]. This can cause simulations to crash. "
            "Check log file for more details."
        )


HEADERS = {
    "^ATOMS": """ """,
    "^BOND": """!
!V(bond) = Kb(b - b0)**2
!
!Kb: kcal/mole/A**2
!b0: A
!
!atom type Kb          b0
!
""",
    "^ANGL": """!
!V(angle) = Ktheta(Theta - Theta0)**2
!
!V(Urey-Bradley) = Kub(S - S0)**2
!
!Ktheta: kcal/mole/rad**2
!Theta0: degrees
!Kub: kcal/mole/A**2 (Urey-Bradley)
!S0: A
!
!atom types     Ktheta    Theta0   Kub     S0
!
""",
    "^DIHE": """!
!V(dihedral) = Kchi(1 + cos(n(chi) - delta))
!
!Kchi: kcal/mole
!n: multiplicity
!delta: degrees
!
!atom types             Kchi    n   delta
!
""",
    "^IMPR": """!
!V(improper) = Kpsi(psi - psi0)**2
!
!Kpsi: kcal/mole/rad**2
!psi0: degrees
!note that the second column of numbers (0) is ignored
!
!atom types           Kpsi                   psi0
!
""",
    "^CMAP": """! 2D grid correction data. 
! Finalfix3, Feig/Best/MacKerell 2010

! Jing Huang/Alex MacKerell adjustments to correct for 
! oversampling of alpha L conformation.  2016/1
""",
    "^NONB": """cutnb 14.0 ctofnb 12.0 ctonnb 10.0 eps 1.0 e14fac 1.0 wmin 1.5 
                !adm jr., 5/08/91, suggested cutoff scheme
!
!V(Lennard-Jones) = Eps,i,j[(Rmin,i,j/ri,j)**12 - 2(Rmin,i,j/ri,j)**6]
!
!epsilon: kcal/mole, Eps,i,j = sqrt(eps,i * eps,j)
!Rmin/2: A, Rmin,i,j = Rmin/2,i + Rmin/2,j
!
!atom  ignored    epsilon      Rmin/2   ignored   eps,1-4       Rmin/2,1-4
!
""",
    "^NBON": """cutnb 14.0 ctofnb 12.0 ctonnb 10.0 eps 1.0 e14fac 1.0 wmin 1.5 
                !adm jr., 5/08/91, suggested cutoff scheme
!
!V(Lennard-Jones) = Eps,i,j[(Rmin,i,j/ri,j)**12 - 2(Rmin,i,j/ri,j)**6]
!
!epsilon: kcal/mole, Eps,i,j = sqrt(eps,i * eps,j)
!Rmin/2: A, Rmin,i,j = Rmin/2,i + Rmin/2,j
!
!atom  ignored    epsilon      Rmin/2   ignored   eps,1-4       Rmin/2,1-4
!
""",
    "^HBON": """! READ PARAM APPEND CARD
! to append hbond parameters from the file: par_hbond.inp
""",
    "^NBFI": """!               Emin        Rmin
!            (kcal/mol)     (A)
!
""",
}


def cleanup_charmm_prm(mol, parameters, outfile):
    import re

    def _is_number(x):
        try:
            float(x)
            return True
        except ValueError:
            return False

    sections = (
        "^ATOMS",
        "^BOND",
        "^ANGL",
        "^DIHE",
        "^THET",
        "^IMPR",
        "^IMPH",
        "^CMAP",
        "^NBON",
        "^NONB",
        "^NBFI",
        "^HBON",
    )

    atom_masses = {
        at.upper(): mol.masses[mol.atomtype == at][0] for at in set(mol.atomtype)
    }
    atom_masses["X"] = 0

    in_section = None

    with open(parameters, encoding="utf-8") as f:
        lines = f.readlines()

    newlines = []
    newlines.append("ATOMS\n")
    for at, mass in sorted(atom_masses.items(), key=lambda x: x[0]):
        newlines.append(f"MASS  -1  {at: >5} {mass:>8.5f}\n")
    newlines.append("\n")

    wrote_newline = False
    for line in lines:
        line = line.strip()
        if line.startswith("*") or line.startswith("!"):
            continue
        if line.startswith("ATOMS") or line.startswith("MASS"):
            continue

        if len(line) == 0:
            if not wrote_newline:
                newlines.append("\n")
            wrote_newline = True
            continue

        # Get stuff before inline comments
        pre_comment = line.split("!")[0].strip()
        if len(pre_comment) == 0:
            continue

        is_section = [re.search(s, pre_comment) is not None for s in sections]
        if any(is_section):
            newlines.append(f"\n{line}\n")
            newlines.append(HEADERS[sections[is_section.index(True)]])
            in_section = sections[is_section.index(True)]
            wrote_newline = False
            continue

        non_numbers = [pp.strip() for pp in pre_comment.split() if not _is_number(pp)]
        types_in_system = [pp.upper() in atom_masses for pp in non_numbers]
        type_line = len(types_in_system)
        if type_line and not all(types_in_system):
            if in_section == "^CMAP":
                skip_cmap_terms = True
            continue
        if type_line and all(types_in_system) and in_section == "^CMAP":
            skip_cmap_terms = False
        if in_section == "^CMAP" and skip_cmap_terms and not type_line:
            continue
        newlines.append(f"{line}\n")
        wrote_newline = False

    with open(outfile, "w", encoding="utf-8") as f:
        f.writelines(newlines)


class _TestCharmmBuild(unittest.TestCase):
    @unittest.skipUnless(_psfgen_exists, "Requires psfgen")
    def test_build(self):
        from moleculekit.molecule import Molecule
        from htmd.builder.solvate import solvate
        from htmd.home import home
        from htmd.util import tempname, assertSameAsReferenceDir
        import os
        import numpy as np

        # Use pre-prepared files so we can tell whether the error is in prepare or in build
        # Inputs are reference outputs of proteinprepare.
        preparedInputDir = home(dataDir="test-proteinprepare")

        pdbids = ["3PTB", "1A25", "1GZM", "1U5U"]
        for pdb in pdbids:
            with self.subTest(pdb=pdb):
                print("Building {}".format(pdb))
                inFile = os.path.join(
                    preparedInputDir, pdb, "{}-prepared.pdb".format(pdb)
                )
                mol = Molecule(inFile)
                mol.filter("protein")  # Fix for bad proteinPrepare hydrogen placing

                np.random.seed(1)  # Needed for ions
                smol = solvate(mol)
                topos = ["top/top_all36_prot.rtf", "top/top_water_ions.rtf"]
                params = ["par/par_all36m_prot.prm", "par/par_water_ions.prm"]
                tmpdir = tempname()
                _ = build(smol, topo=topos, param=params, outdir=tmpdir)

                compareDir = home(dataDir=os.path.join("test-charmm-build", pdb))
                assertSameAsReferenceDir(compareDir, tmpdir)

                # shutil.rmtree(tmpdir)

    @unittest.skipUnless(_psfgen_exists, "Requires psfgen")
    def test_customDisulfideBonds(self):
        from moleculekit.molecule import Molecule
        from htmd.builder.solvate import solvate
        from htmd.home import home
        from htmd.util import tempname, assertSameAsReferenceDir
        import os
        import numpy as np

        # Use pre-prepared files so we can tell whether the error is in prepare or in build
        # Inputs are reference outputs of proteinprepare.
        preparedInputDir = home(dataDir="test-proteinprepare")

        pdb = "1GZM"
        inFile = os.path.join(preparedInputDir, pdb, "{}-prepared.pdb".format(pdb))
        mol = Molecule(inFile)
        mol.filter("protein")  # Fix for bad proteinPrepare hydrogen placing

        np.random.seed(1)  # Needed for ions
        smol = solvate(mol)
        topos = ["top/top_all36_prot.rtf", "top/top_water_ions.rtf"]
        params = ["par/par_all36m_prot.prm", "par/par_water_ions.prm"]
        disu = [
            ["segid 1 and resid 110", "segid 1 and resid 187"],
            ["segid 0 and resid 110", "segid 0 and resid 187"],
        ]
        tmpdir = tempname()
        _ = build(smol, topo=topos, param=params, outdir=tmpdir, disulfide=disu)

        compareDir = home(dataDir=os.path.join("test-charmm-build", pdb))
        assertSameAsReferenceDir(compareDir, tmpdir)

    @unittest.skipUnless(_psfgen_exists, "Requires psfgen")
    def test_disulfideWithInsertion(self):
        from moleculekit.molecule import Molecule
        from htmd.builder.solvate import solvate
        from htmd.home import home
        from htmd.util import tempname, assertSameAsReferenceDir
        import os
        import numpy as np

        # Use pre-prepared files so we can tell whether the error is in prepare or in build
        # Inputs are reference outputs of proteinprepare.
        preparedInputDir = home(dataDir="test-proteinprepare")

        pdb = "3PTB"

        print("Building {}".format(pdb))
        inFile = os.path.join(preparedInputDir, pdb, "{}-prepared.pdb".format(pdb))
        mol = Molecule(inFile)
        mol.filter("protein")  # Fix for bad proteinPrepare hydrogen placing

        np.random.seed(1)  # Needed for ions
        smol = solvate(mol)
        topos = ["top/top_all36_prot.rtf", "top/top_water_ions.rtf"]
        params = ["par/par_all36m_prot.prm", "par/par_water_ions.prm"]

        smol.insertion[smol.resid == 42] = (
            "A"  # Adding an insertion to test that disulfide bonds with insertions work
        )
        tmpdir = tempname()
        _ = build(smol, topo=topos, param=params, outdir=tmpdir)
        compareDir = home(dataDir=os.path.join("test-charmm-build", "3PTB_insertion"))
        assertSameAsReferenceDir(compareDir, tmpdir)


if __name__ == "__main__":
    import unittest

    unittest.main()



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from pint import UnitRegistry
import numpy as np


def convert(fromunit, tounit, value, fstep=1, timestep=4):
    """Converts values between units

    Parameters
    ----------
    fromunit : str
        Unit to convert from
    tounit : str
        Unit to convert to
    value : scalar
        The value to convert
    fstep : float
        The sampling frame step of the simulation in nanoseconds
    timestep : int
        The timestep of the simulation in femtoseconds

    Returns
    -------
    conv : scalra
        The converted value
    """

    ureg = UnitRegistry()
    ureg.define("frame = {} * ns".format(fstep))
    ureg.define("step = ({} / 1000000) * ns = timestep".format(timestep))

    q = ureg.Quantity(value, fromunit)
    convval = q.to(tounit)
    if convval.units == "frame" or convval.units == "step":
        vals = np.round(convval.magnitude).astype(int)
        if vals.size == 1:  # Fix for tica. remove in future
            return int(vals)
        return vals
    else:
        return convval.magnitude



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

# Rapidly exploring random tree to find reaction coordinate for umbrella sampling!
# Use VMD draw line to visualize! (dashed for failed, solid for connected)
import numpy as np
from moleculekit.util import maxDistance, uniformRandomRotation
from moleculekit.molecule import Molecule
from moleculekit.vmdviewer import getCurrentViewer
from scipy.spatial.distance import cdist


class Tree:
    def __init__(self, initialpoints):
        if isinstance(initialpoints, np.ndarray):
            initialpoints = initialpoints.tolist()
        self.points = initialpoints
        self.parent = [None] * len(self.points)
        self.costs = [0] * len(self.points)

    def addPoint(self, point, parent, dist):
        if isinstance(point, np.ndarray):
            point = point.tolist()
        if isinstance(dist, np.ndarray):
            dist = dist[0]
        self.points.append(point)
        self.parent.append(parent)
        self.costs.append(self.costs[parent] + dist)


def _getCoordinates(mol: Molecule, ligandsel, ligcom, othersel="protein"):
    othersel = f"({othersel}) and not ({ligandsel})"
    translation = mol.center(sel=othersel)
    othercoor = mol.get("coords", sel=othersel)
    if ligcom:
        ligcoor = np.mean(mol.get("coords", sel=ligandsel), axis=0)[np.newaxis, :]
    else:
        ligcoor = mol.get("coords", sel=ligandsel)
    return othercoor, ligcoor, translation


def _newPoint(p_target, p_begin, step):
    v = p_target - p_begin
    norm_v = v / np.linalg.norm(v)
    if isinstance(step, np.ndarray):
        return p_begin + (step[:, np.newaxis] * norm_v)
    else:
        return p_begin + (step * norm_v)


def _dist(a, b):
    if isinstance(a, list):
        a = np.array(a)
    if isinstance(b, list):
        b = np.array(b)
    return cdist(np.atleast_2d(a), np.atleast_2d(b))


def _getNearest(coords, point):
    dist = _dist(coords, point)
    idx = np.argmin(dist)
    return idx, coords[idx]


def _collision(coords, point, buffer=2):
    dist = _dist(coords, point)
    if np.any(dist < buffer):
        return True
    else:
        return False


def _pointsOnSphere(radius, numsamples=1000):
    pointcoords = np.zeros((numsamples, 3))
    for i in range(numsamples):
        point = Molecule()
        point.empty(1)
        point.coords = np.zeros((1, 3, 1), dtype=np.float32)
        point.moveBy([0, 0, radius])
        point.rotateBy(uniformRandomRotation())
        pointcoords[i, :] = np.squeeze(point.coords)

    return pointcoords


def _collisionFreePath(obstacles, start, stop, colldist, step):
    linepoints = _newPoint(stop, start, np.arange(step, _dist(start, stop), step=step))
    distances = _dist(linepoints, obstacles)
    if np.any(distances <= colldist):
        return False
    else:
        return True


def _collisionFreeNeighbours(tree, point, radius, obstacles, colldist, step):
    dists = _dist(tree.points, point)
    near = np.where(dists < radius)[0]
    nn = []
    for n in near:
        if _collisionFreePath(obstacles, point, tree.points[n], colldist, step):
            nn.append(n)
    near = nn
    return near, dists[near]


def _rewire(tree, near, neardist, newidx):
    # Remove parent from near neighbours
    parentidx = np.where(near == tree.parent[newidx])[0]
    near = np.delete(near, parentidx)
    neardist = np.delete(neardist, parentidx)
    if len(near) == 0:
        return

    costs = np.array(tree.costs)
    newcosts = np.squeeze(neardist) + costs[newidx]
    shorter = np.where(newcosts < costs[near])[0]
    for s in shorter:
        tree.costs[near[s]] = newcosts[s]
        tree.parent[near[s]] = newidx


def _chooseParent(tree, near, neardist):
    if len(near) == 0:
        raise AssertionError("Should include at least p_near.")

    costs = np.array(tree.costs)
    closecosts = costs[near] + np.squeeze(neardist)
    shortest = np.argmin(closecosts)
    return near[shortest], neardist[shortest]


def _endCondition(protcoor, target, curr, outdist, method="target"):
    if method == "target":
        return _dist(target, curr) < 2
    if method == "exited":
        return np.all(_dist(protcoor, curr) >= outdist)


def _pathOptimize(tree, currnode, obstacles, colldist, step):
    start = np.array(tree.points[currnode])
    endnode = tree.parent[currnode]
    prevend = endnode

    while True:
        if endnode is None:
            tree.parent[currnode] = prevend
            tree.costs[currnode] = tree.costs[prevend] + _dist(
                tree.points[currnode], tree.points[prevend]
            )
            return

        stop = np.array(tree.points[endnode])
        if _collisionFreePath(obstacles, start, stop, colldist, step):
            prevend = endnode
            endnode = tree.parent[endnode]
        else:
            tree.parent[currnode] = prevend
            tree.costs[currnode] = tree.costs[prevend] + _dist(
                tree.points[currnode], tree.points[prevend]
            )
            _pathOptimize(tree, prevend, obstacles, colldist, step)
            return


def _getBeacons(tree, endnode):
    beacons = []
    currnode = tree.parent[endnode]
    while tree.parent[currnode] is not None:
        beacons.append(currnode)
        currnode = tree.parent[currnode]
    return beacons


def _randomPoint(min, max):
    lens = max - min
    return np.random.rand(3) * lens + min


def rrtstarsmart(
    mol,
    ligandsel,
    step=1,
    maxiter=int(1e6),
    ligcom=False,
    colldist=2,
    outdist=8,
    radius=2.5,
):
    protcoor, ligcoor, translation = _getCoordinates(mol, ligandsel, ligcom)
    viewer = _prepareViewer(mol, ligandsel)

    mincoor = np.squeeze(np.min(mol.coords, axis=0) - 10)
    maxcoor = np.squeeze(np.max(mol.coords, axis=0) + 10)
    print(mincoor, maxcoor)

    tree = Tree(ligcoor)

    for i in range(maxiter):
        print(i)
        # p_rand = spherecoor[np.random.randint(spherecoor.shape[0]), :]
        p_rand = _randomPoint(mincoor, maxcoor)
        parent_idx, p_near = _getNearest(tree.points, p_rand)
        p_new = _newPoint(p_rand, p_near, step)

        if _collision(protcoor, p_new, buffer=colldist):
            continue

        print("no collision")
        near, neardist = _collisionFreeNeighbours(
            tree, p_new, radius, protcoor, colldist, step
        )
        parent_idx, dist = _chooseParent(tree, near, neardist)
        tree.addPoint(p_new, parent_idx, dist)
        _rewire(tree, near, neardist, len(tree.points) - 1)

        if _endCondition(protcoor, p_rand, p_new, outdist, method="exited"):
            break

    for i in range(len(tree.points)):
        if tree.parent[i] is not None:
            _drawline(viewer, tree.points[i], tree.points[tree.parent[i]])

    print("optimizing")
    _pathOptimize(tree, len(tree.points) - 1, protcoor, colldist, step)
    print("finished optimization")
    # TODO: The algorithm is not finished. Normally you would get beacons and continue sampling around them to optimize
    # TODO: the path even more. However, we don't really need this in our case, unless I decide to make a real full impl
    # TODO: http://ieeexplore.ieee.org/xpls/icp.jsp?arnumber=6284384

    currnode = len(tree.points) - 1
    viewer.send("draw color green")
    while tree.parent[currnode] is not None:
        _drawline(viewer, tree.points[currnode], tree.points[tree.parent[currnode]])
        currnode = tree.parent[currnode]

    return


def rrt(mol, ligandsel, step=1, maxiter=int(1e6), ligcom=False, colldist=2, outdist=8):
    protcoor, ligcoor, translation = _getCoordinates(mol, ligandsel, ligcom)
    viewer = _prepareViewer(mol, ligandsel)

    spherecoor = _pointsOnSphere(maxDistance(mol) + 5)

    tree = Tree(ligcoor)
    nocol = None

    for i in range(maxiter):
        print(i)
        if nocol is None:
            p_rand = spherecoor[
                np.random.randint(spherecoor.shape[0]), :
            ]  # TODO: Make it clever. Discard failed points
        else:  # Reusing the same point that caused no collision
            p_rand = nocol
        parent_idx, p_near = _getNearest(tree.points, p_rand)
        p_new = _newPoint(p_rand, p_near, step)

        if _collision(protcoor, p_new, buffer=colldist):
            nocol = None
            continue
        nocol = p_rand

        print("no collision")
        tree.addPoint(p_new, parent_idx, step)

        _drawline(viewer, p_near, p_new)
        if _dist(p_rand, p_new) < 2:
            break


def raytracing(
    mol,
    ligandsel,
    othersel="protein",
    step=1,
    colldist=2,
    outdist=8,
    ligcom=False,
    numsamples=2000,
    ratioexposed=0,
    vmd=True,
):
    """Find the escape vector of a ligand from a pocket.

    This function creates a sphere of points around the `othersel` atoms and traces a
    line from each atom of the ligand to each point on the sphere. If the line does not
    hit any `othersel` atoms within `colldist` it is checked if it exits the pocket by
    checking if any point on the line is outside the pocket.
    Thus we can calculate all the unimpeded vectors with which the ligand can reach the
    solvent.
    We then select the point of the surrounding sphere to which most ligand atoms can
    reach without hitting any `othersel` atoms within `colldist` and we consider this
    as the escape vector.

    Parameters
    ----------
    mol : Molecule
        The molecule to analyze.
    ligandsel : str
        The selection string to use to identify the ligand.
    othersel : str, optional
        The selection string to use to identify the other molecules which impede the
        escape of the ligand.
    step : float, optional
        The step size to use for the line tracing.
    colldist : float, optional
        The collision distance threshold to use for the line tracing.
    outdist : float, optional
        The distance threshold to use to check if the line exits the pocket.
    ligcom : bool, optional
        Whether to use the center of mass of the ligand or its individual atoms to
        trace the line.
    numsamples : int, optional
        The number of points to sample on the surrounding sphere.
    ratioexposed : float, optional
        The ratio of exposed ligand atoms that should be present on the escape vector.
        If set to a value greater than 0, the function will check if the number of
        exposed ligand atoms on the escape vector is greater than the specified ratio
        and raise an error if it is not.
    vmd : bool, optional
        Whether to visualize the results using VMD.

    Returns
    -------
    translation : np.ndarray
        The translation vector to apply if we want to align the molecule to the
        escape vector.
    escape_vector : np.ndarray
        The escape vector, i.e. the vector which the ligand can use to reach the
        solvent without hitting any `othersel` atoms within `colldist`.

    Examples
    --------
    >>> from moleculekit.util import rotation_matrix_from_vectors
    >>> translation, escape_vector = raytracing(mol, "resname LIG", "protein")
    >>> rotmat = rotation_matrix_from_vectors(np.array([0, 0, 1]), escape_vector) # align z-axis with escape vector
    >>> mol.moveBy(translation)
    >>> mol.rotateBy(rotmat)
    """
    from tqdm import tqdm

    mol = mol.copy()
    protcoor, ligcoor, translation = _getCoordinates(mol, ligandsel, ligcom, othersel)

    spherecoor = _pointsOnSphere(maxDistance(mol) + 5, numsamples=numsamples)
    if vmd:
        _viewSphere(spherecoor)
        viewer = _prepareViewer(mol, ligandsel)

    distances = _dist(ligcoor, spherecoor)

    results = []
    for j in tqdm(range(spherecoor.shape[0]), desc="Searching for pathway"):
        results.append(
            parallelfunc(
                j,
                spherecoor[j, :],
                ligcoor,
                protcoor,
                step,
                colldist,
                outdist,
                distances[:, j],
            )
        )

    points = []
    pointdist = []
    shortpoints = []
    for r in results:
        points += r[0]
        pointdist += r[1]
        shortpoints += r[2]

    if len(points) == 0:
        raise RuntimeError(
            f"No ligand atoms can exit the pocket without {colldist}A clashes."
        )

    points = np.array(points)
    numexposed = len(np.unique(points[:, 0]))
    percentexposed = (numexposed / ligcoor.shape[0]) * 100
    if numexposed < (ratioexposed * ligcoor.shape[0]):
        raise RuntimeError(
            f"Only {percentexposed:.1f}% ligand atoms can exit the pocket without {colldist}A clashes. "
            f"This collides with the user-defined required {ratioexposed * 100:.1f}% exposed ligand atoms."
        )
    print(
        f"{percentexposed:.1f}% ligand atoms can exit the pocket without {colldist}A clashes."
    )

    from scipy.stats import mode

    modesphere = mode(points[:, 1], keepdims=False).mode
    idx = points[:, 1] == modesphere
    meanlig = np.mean(ligcoor[points[idx, 0], :], axis=0)

    if vmd:
        viewer.send("draw color green")
        _drawline(viewer, meanlig, spherecoor[modesphere, :])

        for idx, p in enumerate(points.tolist()):
            if p[1] == modesphere:
                viewer.send("draw color red")
                _drawline(viewer, shortpoints[idx], spherecoor[p[1], :])
                viewer.send("draw color green")
                _drawline(viewer, ligcoor[p[0], :], shortpoints[idx])

    return translation, spherecoor[modesphere, :] - meanlig


def parallelfunc(j, spherecoor, ligcoor, protcoor, step, colldist, outdist, distances):
    points = []
    pointdist = []
    shortpoints = []
    for i in range(ligcoor.shape[0]):
        linepoints = _newPoint(
            spherecoor, ligcoor[i, :], np.arange(step, distances[i], step=step)
        )
        collisions = _dist(linepoints, protcoor)
        if not np.any(collisions <= colldist):
            linep_mincoll = np.min(collisions, axis=1)
            idx = np.where(linep_mincoll >= outdist)[0]
            if len(idx) == 0:
                continue
            min_idx = np.min(idx)
            points.append([i, j])
            pointdist.append(step * (min_idx + 1))
            shortpoints.append(linepoints[min_idx, :])
    return points, pointdist, shortpoints


def _drawline(viewer, start, end, dashed=True):
    if dashed:
        append = "style dashed"
    else:
        append = ""
    viewer.send(
        "draw line {{ {} }} {{ {} }} {}".format(
            " ".join(map(str, start)), " ".join(map(str, end)), append
        )
    )


def _prepareViewer(mol, ligandsel):
    mol.view(sel="protein", style="lines", hold=True, viewer="vmd")
    mol.view(sel=ligandsel, style="licorice", hold=True, viewer="vmd")
    mol.view(viewer="vmd")

    viewer = getCurrentViewer()
    viewer.send("draw color red")
    viewer.send("color Display Background black")
    viewer.send("draw materials off")
    return viewer


def _viewSphere(spherecoor):
    spheremol = Molecule()
    spheremol.empty(spherecoor.shape[0])
    spheremol.coords = np.atleast_3d(spherecoor)
    spheremol.view(guessBonds=False, viewer="vmd")
